import "Turbine";
import "Turbine.Gameplay";
import "Turbine.UI";
import "Turbine.UI.Lotro";
import "RobYap.CleanChatWindows.Libraries.Hint"
import "RobYap.CleanChatWindows.Libraries.Text"
import "RobYap.CleanChatWindows.Libraries.MessageBox"

local commonDir = "Thurallor.Common";
import (commonDir .. ".Turbine");
import (commonDir .. ".Utils.Utils_11");
import (commonDir .. ".Utils.Locale_3");
import (commonDir .. ".Utils.Color_1");
import (commonDir .. ".Utils.AutoSizing_7");
import (commonDir .. ".UI.RadioButton_2");
import (commonDir .. ".UI.ColorPicker_5");

---@diagnostic disable-next-line: undefined-global
CleanChatWindow = class(Turbine.UI.Window);

function CleanChatWindow:Constructor(settings)

  ------------------------------------------BEGIN SETTINGS AND LOCALS INIT------------------------------------------------------------------

  self._settings = settings;

  if settings == nil then self._settings = {} end

  if true then
    if self._settings.Name              == nil
        or self._settings.Name          == ""  then error("Name required") end
    if self._settings.Width             == nil then self._settings.Width              = 350   end
    if self._settings.Height            == nil then self._settings.Height             = 250   end
    if self._settings.Left              == nil then self._settings.Left               = (Turbine.UI.Display.GetWidth() / 2) - (self._settings.Width / 2) end
    if self._settings.Top               == nil then self._settings.Top                = (Turbine.UI.Display.GetHeight() / 2) - (self._settings.Height / 2) end
    if self._settings.CollapseRight     == nil then self._settings.CollapseRight      = false end
    if self._settings.Active            == nil then self._settings.Active             = true  end
    if self._settings.ToolbarBottom     == nil then self._settings.ToolbarBottom      = false end
    if self._settings.Enabled           == nil then self._settings.Enabled            = true  end
    if self._settings.TipsEnabled       == nil then self._settings.TipsEnabled        = true  end
    if self._settings.Maximized         == nil then self._settings.Maximized          = true  end
    if self._settings.ChatClickThrough  == nil then self._settings.ChatClickThrough   = true  end
    if self._settings.InteractiveMode   == nil then self._settings.InteractiveMode    = false end
    if self._settings.ToolbarPinned     == nil then self._settings.ToolbarPinned      = true end
    if self._settings.HoverActivate     == nil then self._settings.HoverActivate      = true end
    if self._settings.WindowOpacity     == nil then self._settings.WindowOpacity      = 1.00  end
    if self._settings.ChatOpacity       == nil then self._settings.ChatOpacity        = 0.75  end
    if self._settings.ChatHoverOpacity  == nil then self._settings.ChatHoverOpacity   = 0.75  end
    if self._settings.ToolbarOpacity    == nil then self._settings.ToolbarOpacity     = 0.75  end
    if self._settings.ActiveChatGroupId == nil then self._settings.ActiveChatGroupId  = -1    end
    if self._settings.PurgeChatMinutes  == nil then self._settings.PurgeChatMinutes   = 30 end
    if self._settings.WindowBackColor   == nil then self._settings.WindowBackColor    = Turbine.UI.Color( 0, 0, 0) end
    if self._settings.WindowForeColor   == nil then self._settings.WindowForeColor    = Turbine.UI.Color( 1, 1, 1 )  end
    if self._settings.ToolbarBackColor  == nil then self._settings.ToolbarBackColor   = Turbine.UI.Color( 0, 0, 0) end
    if self._settings.ToolbarForeColor  == nil then self._settings.ToolbarForeColor   = Turbine.UI.Color( 0.7, 0.7, 0.7) end
    if self._settings.ToggleColor       == nil then self._settings.ToggleColor        = Turbine.UI.Color( 1, 0.62, 0.22 ) end
    if self._settings.WarningColor      == nil then self._settings.WarningColor       = Turbine.UI.Color( 0.59, 0.14, 0.17 ) end
    if self._settings.DisabledColor     == nil then self._settings.DisabledColor      = Turbine.UI.Color( 0.92, 0.21, 0.21 )  end

    if self._settings.ChatHoverOpacity < self._settings.ChatOpacity then
      self._settings.ChatHoverOpacity   = self._settings.ChatOpacity;
    end

    if self._settings.ChatGroups == nil then
      self._settings.ChatGroups = {};
      table.insert(self._settings.ChatGroups,
                  {
                      Id                = 1,
                      Order             = 1,
                      Name              = "Default",
                      IncludeTimestamps = false,
                      SelfLootNoCoins   = false,
                      KinNoAnnounce     = false,
                      Channels          = {},
                      ChannelColors     = {}
                  }
      );
    end

  end

  self._chatGroupCount = 0;
  self._chatGroupMaxId = 0;
  self._chatGroupMaxOrder = 0;

  for key,value in spairs(self._settings.ChatGroups, function(t,aa,bb) return t[bb].Order > t[aa].Order end) do
    self._settings.ChatGroups[key].FirstMessageTime = -1;
    self._settings.ChatGroups[key].PurgeCount       = -1;
    self._settings.ChatGroups[key].MessageCount     = -1;
  end

  self.UpdateChatGroups = function()

    self._chatGroupCount = 0;
    self._chatGroupMaxId = 0;
    self._chatGroupMaxOrder = 0;

    for key,value in spairs(self._settings.ChatGroups, function(t,aa,bb) return t[bb].Order > t[aa].Order end) do
      if self._settings.ActiveChatGroupId == -1 then
        self._settings.ActiveChatGroupId = self._settings.ChatGroups[key].Id;
      end
      --DebugWrite("self._settings.ActiveChatGroupId (A): " .. tostring(self._settings.ActiveChatGroupId));
      if self._settings.ChatGroups[key].Id > self._chatGroupMaxId then
        self._chatGroupMaxId = self._settings.ChatGroups[key].Id;
      end
      self._chatGroupMaxOrder = self._settings.ChatGroups[key].Order;
    end


    for key,value in spairs(self._settings.ChatGroups, function(t,aa,bb) return t[bb].Order > t[aa].Order end) do
      --DebugWrite("key: " .. tostring(key));
      self._chatGroupCount = self._chatGroupCount + 1;

      if self._settings.ChatGroups[key].Id == nil then
        self._chatGroupMaxId = self._chatGroupMaxId + 1;
        self._settings.ChatGroups[key].Id = self._chatGroupMaxId;
      end
      if self._settings.ChatGroups[key].Order == nil then
        self._chatGroupMaxOrder  = self._chatGroupMaxOrder + 1;
        self._settings.ChatGroups[key].Order = self._chatGroupMaxOrder;
      end
      if self._settings.ChatGroups[key].Name              == nil then self._settings.ChatGroups[key].Name               = "Default" end
      if self._settings.ChatGroups[key].IncludeTimestamps == nil then self._settings.ChatGroups[key].IncludeTimestamps  = false end
      if self._settings.ChatGroups[key].SelfLootNoCoins   == nil then self._settings.ChatGroups[key].SelfLootNoCoins    = false end
      if self._settings.ChatGroups[key].KinNoMOTD         == nil then self._settings.ChatGroups[key].KinNoMOTD          = false end
      if self._settings.ChatGroups[key].KinNoAnnounce     == nil then self._settings.ChatGroups[key].KinNoAnnounce      = false end

      if self._settings.ChatGroups[key].FirstMessageTime  == nil then self._settings.ChatGroups[key].FirstMessageTime   = -1 end
      if self._settings.ChatGroups[key].PurgeCount        == nil then self._settings.ChatGroups[key].PurgeCount         = -1 end
      if self._settings.ChatGroups[key].MessageCount      == nil then self._settings.ChatGroups[key].MessageCount       = 0 end

      -- Custom type of 400 is /rolls only
      -- Custom type of 500 is Non-Player say
      if self._settings.ChatGroups[key].Channels                                    == nil then self._settings.ChatGroups[key].Channels = {} end

      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.Advancement]      == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.Advancement]  = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.Death]            == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.Death]        = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.Emote]            == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.Emote]        = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.EnemyCombat]      == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.EnemyCombat]  = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.Error]            == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.Error]        = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.FellowLoot]       == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.FellowLoot]   = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.Fellowship]       == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.Fellowship]   = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.Kinship]          == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.Kinship]      = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.LFF]              == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.LFF]          = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.OOC]              == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.OOC]          = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.Officer]          == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.Officer]      = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.PlayerCombat]     == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.PlayerCombat] = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.Quest]            == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.Quest]        = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.Raid]             == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.Raid]         = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.Regional]         == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.Regional]     = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.Roleplay]         == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.Roleplay]     = false end
      -- Custom: Player Say
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.Say]              == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.Say]          = true end
      -- Custom: Non-Player Say
      if self._settings.ChatGroups[key].Channels[_G.CHATTYPE_NPC_Say]               == nil then self._settings.ChatGroups[key].Channels[_G.CHATTYPE_NPC_Say]           = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.SelfLoot]         == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.SelfLoot]     = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.Standard]         == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.Standard]     = false end
      -- Custom: Rolls only
      if self._settings.ChatGroups[key].Channels[_G.CHATTYPE_Rolls]                 == nil then self._settings.ChatGroups[key].Channels[_G.CHATTYPE_Rolls]             = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.Tell]             == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.Tell]         = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.Trade]            == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.Trade]        = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.Unfiltered]       == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.Unfiltered]   = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.UserChat1]        == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.UserChat1]    = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.UserChat2]        == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.UserChat2]    = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.UserChat3]        == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.UserChat3]    = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.UserChat4]        == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.UserChat4]    = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.UserChat5]        == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.UserChat5]    = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.UserChat6]        == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.UserChat6]    = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.UserChat7]        == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.UserChat7]    = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.UserChat8]        == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.UserChat8]    = false end
      if self._settings.ChatGroups[key].Channels[Turbine.ChatType.World]            == nil then self._settings.ChatGroups[key].Channels[Turbine.ChatType.World]        = false end
      -- Not sure : Event broadcast?
      if self._settings.ChatGroups[key].Channels[_G.CHATTYPE_EventBroadcast]        == nil then self._settings.ChatGroups[key].Channels[_G.CHATTYPE_EventBroadcast]    = false end


      if self._settings.ChatGroups[key].ChannelColors                == nil then self._settings.ChatGroups[key].ChannelColors = {} end

      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Advancement]  == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Advancement] = "#005ebd" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Death]        == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Death] = "#b30101" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Emote]        == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Emote] = "#7e7e7e" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.EnemyCombat]  == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.EnemyCombat] = "#ff0000" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Error]        == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Error] = "#d2d2d2" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.FellowLoot]   == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.FellowLoot] = "#00b545" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Fellowship]   == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Fellowship] = "#42a830" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Kinship]      == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Kinship] = "#52adff" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.LFF]          == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.LFF] = "#d2d2d2" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.OOC]          == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.OOC] = "#d2d2d2" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Officer]      == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Officer] = "#4b9ee9" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.PlayerCombat] == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.PlayerCombat] = "#f7d200" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Quest]        == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Quest] = "#d2d2d2" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Raid]         == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Raid] = "#d2d2d2" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Regional]     == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Regional] = "#d2d2d2" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Roleplay]     == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Roleplay] = "#d2d2d2" end
      -- Custom: Player Say
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Say]          == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Say] = "#ffffff" end
      -- Custom: Non-Player Say
      if self._settings.ChatGroups[key].ChannelColors[_G.CHATTYPE_NPC_Say]           == nil then self._settings.ChatGroups[key].ChannelColors[_G.CHATTYPE_NPC_Say] = "#d2d2d2" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.SelfLoot]     == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.SelfLoot] = "#e59d9d" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Standard]     == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Standard] = "#c0b566" end
      -- Custom: Rolls only
      if self._settings.ChatGroups[key].ChannelColors[_G.CHATTYPE_Rolls]             == nil then self._settings.ChatGroups[key].ChannelColors[_G.CHATTYPE_Rolls] = "#c0b566" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Tell]         == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Tell] = "#e6c74d" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Trade]        == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Trade] = "#d2d2d2" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Unfiltered]   == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.Unfiltered] = "#d2d2d2" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.UserChat1]    == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.UserChat1] = "#d2d2d2" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.UserChat2]    == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.UserChat2] = "#d2d2d2" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.UserChat3]    == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.UserChat3] = "#d2d2d2" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.UserChat4]    == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.UserChat4] = "#d2d2d2" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.UserChat5]    == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.UserChat5] = "#d2d2d2" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.UserChat6]    == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.UserChat6] = "#d2d2d2" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.UserChat7]    == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.UserChat7] = "#d2d2d2" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.UserChat8]    == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.UserChat8] = "#d2d2d2" end
      if self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.World]        == nil then self._settings.ChatGroups[key].ChannelColors[Turbine.ChatType.World] = "#d2d2d2" end
      -- Not sure : Event broadcast?
      if self._settings.ChatGroups[key].ChannelColors[_G.CHATTYPE_EventBroadcast]    == nil then self._settings.ChatGroups[key].ChannelColors[_G.CHATTYPE_EventBroadcast] = "#fb9000" end
    end

    if self.ChatGroup ~= nil then
      self.ChatGroup:UpdateChatGroups();
    end

  end

  self:UpdateChatGroups();

  self._headerHeight    = 14;
  self._groupHeight     = 18;
  self._scrollbarWidth  = 10;

  --Right toolbar button start/padding
  self._rightButtonPos  = self._settings.Width;
  --Left toolbar button start/padding
  self._leftButtonPos   = 0; 

  --Used when dragging the window
  self._dragStartX      = 0;
  self._dragStartY      = 0;

  --Used to store message data
  self._entityMessages      = {};
  self._entityTypes         = {};
  self._entityMessageCounts = {};

  ------------------------------------------END SETTINGS AND LOCALS INIT------------------------------------------------------------------



  ------------------------------------------ MAIN WINDOW --------------------------------------------------------

  Turbine.UI.Window.Constructor( self );
  self:SetOpacity(self._settings.WindowOpacity);
  self:SetVisible(self._settings.Active);
  self:SetBackColor(Turbine.UI.Color( self._settings.ChatOpacity, self._settings.WindowBackColor.R, self._settings.WindowBackColor.G, self._settings.WindowBackColor.B ));
  self:SetPosition(self._settings.Left, self._settings.Top);
  self:SetSize(self._settings.Width, self._settings.Height);
  self:SetMouseVisible(false);
  self.HoverActive = false;

  self.HexToColor = function(hex, alpha)
    local redColor,greenColor,blueColor=hex:match('#?(..)(..)(..)')
    redColor, greenColor, blueColor = tonumber("0x"..redColor, 16)/255, tonumber("0x"..greenColor, 16)/255, tonumber("0x"..blueColor, 16)/255
    redColor, greenColor, blueColor = math.floor(redColor*100)/100, math.floor(greenColor*100)/100, math.floor(blueColor*100)/100
    if alpha == nil then
      return redColor, greenColor, blueColor
    end
    return redColor, greenColor, blueColor, alpha
  end

  self.PickColor = function(target)
    ---@diagnostic disable-next-line: undefined-global
    local colorPicker = Thurallor.UI.ColorPicker(target:GetBackColor());

    colorPicker.Accepted = function(picker)
      local color = picker:GetColor();
      target:SetBackColor(color);
      target:ColorUpdated();
    end

    colorPicker.Canceled = function(picker)
    end
  end

  self.PickChatColor = function(sender,args)
    local channelChatColor = self._settings.ChatGroups[args.ChatGroupId].ChannelColors[args.ChatType];
    ---@diagnostic disable-next-line: undefined-global
    local colorPicker = Thurallor.UI.ColorPicker(Turbine.UI.Color(self.HexToColor(channelChatColor, nil)));
    colorPicker.ChatType = args.ChatType;
    colorPicker.ChannelLabel = args.Label;

    colorPicker.Accepted = function(picker)
      local color = picker:GetColor();
      self._settings.ChatGroups[args.ChatGroupId].ChannelColors[picker.ChatType] = "#" .. string.format("%02x", color.R*255) .. string.format("%02x", color.G*255) .. string.format("%02x", color.B*255);
      picker.ChannelLabel:SetBackColor(color);
      Save();
    end

    colorPicker.Canceled = function(picker)
    end
  end

  self.ToggleMinMax = function()

    self._settings.Maximized = not self._settings.Maximized;
    if self._settings.Maximized then

      self:SetPosition(self._settings.Left, self._settings.Top);
      self:SetHeight(self._settings.Height);
      self:SetWidth(self._settings.Width);
      self.ChatWrapper:SetChatWindowsSize(self._settings.Width, self._settings.Height-(self:GetHeaderHeightWithChatGroups()));
    else
      self.Toolbar:SetTop(0);
      self.ToolbarBackground:SetTop(self.Toolbar:GetTop());

      if self._settings.ToolbarBottom then
        self:SetTop((self:GetTop() + self._settings.Height) - self._headerHeight);
      end

      local windowRight = self:GetLeft() + self:GetWidth();
      self:SetWidth(self.MinMax:GetLeft() + self.MinMax:GetWidth() + 1);
      self:SetHeight(self._headerHeight + 1);
      if self._settings.CollapseRight then
        self:SetLeft(windowRight - self:GetWidth());
      end
    end
    self.Toolbar:UpdateLayout();
    self.ChatGroups:UpdateLayout();

    if self._settings.Maximized then
      self.ChatWrapper:SetChatWindowsVisible(true);
    else
      self.ChatWrapper:SetChatWindowsVisible(false);
    end

    self.MinMax:UpdateButton();

    --DebugWrite("self.MinMax.GetTop(): "..tostring(self.MinMax:GetTop()));

  end

  self.UpdateColors = function()

    if self.HoverActive then
      self:SetBackColor(Turbine.UI.Color( self._settings.ChatHoverOpacity, self._settings.WindowBackColor.R, self._settings.WindowBackColor.G, self._settings.WindowBackColor.B ));
    else
      self:SetBackColor(Turbine.UI.Color( self._settings.ChatOpacity, self._settings.WindowBackColor.R, self._settings.WindowBackColor.G, self._settings.WindowBackColor.B ));
    end

    self.ChatWrapper:SetChatWindowsForeColor( Turbine.UI.Color(1, self._settings.WindowForeColor.R, self._settings.WindowForeColor.B, self._settings.WindowForeColor.G) );

    self.ToolbarBackground:SetBackColor(Turbine.UI.Color(self._settings.ToolbarOpacity, self._settings.ToolbarBackColor.R, self._settings.ToolbarBackColor.G, self._settings.ToolbarBackColor.B));
    local buttons = self.Toolbar:GetControls();
    for i=1,buttons:GetCount() do
      if buttons:Get(i).SetForeColor ~= nil then
        ---@diagnostic disable-next-line: undefined-field
        buttons:Get(i):SetForeColor(Turbine.UI.Color(1, self._settings.ToolbarForeColor.R, self._settings.ToolbarForeColor.G, self._settings.ToolbarForeColor.B));
      end
    end

    if self._settings.ChatClickThrough then
      self.Interactive:SetForeColor(Turbine.UI.Color(1, self._settings.ToolbarForeColor.R, self._settings.ToolbarForeColor.G, self._settings.ToolbarForeColor.B));
    else
      self.Interactive:SetForeColor(Turbine.UI.Color(1, self._settings.ToggleColor.R, self._settings.ToggleColor.G, self._settings.ToggleColor.B));
    end

    local textLength = self.ChatWrapper:GetChatWindowsById(self._settings.ActiveChatGroupId):GetTextLength();
    if textLength > 16000 then
      self.Clear:SetForeColor(Turbine.UI.Color(1, self._settings.ToggleColor.R, self._settings.ToggleColor.G, self._settings.ToggleColor.B));
    elseif textLength > 32000 then
      self.Clear:SetForeColor(Turbine.UI.Color(1, self._settings.WarningColor.R, self._settings.WarningColor.G, self._settings.WarningColor.B));
    end

  end

  self.OnHover = function(s,a)
    --DebugWrite("self.HoverActive: " .. tostring(self.HoverActive))
    if not self.HoverActive then
      self.HoverActive = true;
      if self._settings.ChatHoverOpacity >= self._settings.ChatOpacity then
        self:SetBackColor(Turbine.UI.Color( self._settings.ChatHoverOpacity, self._settings.WindowBackColor.R, self._settings.WindowBackColor.G, self._settings.WindowBackColor.B ));
        if self._settings.ChatHoverOpacity > self._settings.ToolbarOpacity then
          self.ToolbarBackground:SetBackColor(Turbine.UI.Color( self._settings.ChatHoverOpacity, self._settings.ToolbarBackColor.R, self._settings.ToolbarBackColor.G, self._settings.ToolbarBackColor.B ));
        end
      end
    end
    if self._settings.HoverActivate then
      self:Activate();
    end
  end

  self.MouseOutOfRange = function()
    local mouseX, mouseY = Turbine.UI.Display.GetMousePosition();

    local top = self:GetTop();
    local bottom = self:GetTop() + self.ChatWrapper:GetTop() + self.ChatWrapper:GetHeight();
    local left = self:GetLeft() + self.ChatWrapper:GetLeft();
    local right = self:GetLeft() + self.ChatWrapper:GetLeft() + self.ChatWrapper:GetWidth();

    if self._settings.ToolbarBottom then
      top = self:GetTop() + self.ChatWrapper:GetTop();
      bottom = self:GetTop() + self:GetHeight();
    end

    if not self._settings.ToolbarPinned then
      bottom = self:GetTop() + self.Toolbar:GetTop() + self.Toolbar:GetHeight();
      if not self._settings.ChatClickThrough then
        if not self._settings.ToolbarBottom then
          top = self:GetTop() + self.Toolbar:GetTop();
          bottom = self:GetTop() + self.ChatWrapper:GetTop() + self.ChatWrapper:GetHeight();
        end
      end
    end

    if self._settings.ChatClickThrough then
      if self._settings.ToolbarBottom then
        top = self:GetTop() + self.Toolbar:GetTop();
      else
        top = self:GetTop();
      end
    end

    if not self._settings.Maximized then
      if self._settings.CollapseRight then
        left = self:GetLeft() + self.MinMax2:GetLeft();
        right = self:GetLeft() + self.MinMax2:GetLeft() + self.MinMax2:GetWidth();
      else
        left = self:GetLeft() + self.MinMax:GetLeft();
        right = self:GetLeft() + self.MinMax:GetLeft() + self.MinMax2:GetWidth();
      end
    end

    if false then
      DebugWrite("Mouse X: " .. tostring(mouseX) .. " Y: " .. tostring(mouseY));
      DebugWrite("  L: " .. tostring(left) .. " T: " .. tostring(top) .. " R: " .. right .. " B: " .. bottom);
      DebugWrite("  mouseX <= left  : " .. tostring(mouseX <= left));
      DebugWrite("  mouseX >= right : " .. tostring(mouseX >= right));
      DebugWrite("  mouseY <= top   : " .. tostring(mouseY <= top));
      DebugWrite("  mouseY >= bottom: " .. tostring(mouseY >= bottom));
    end

    if mouseX <= left
    or mouseX >= right
    or mouseY <= top
    or mouseY >= bottom
    then
      --DebugWrite("MouseOutOfRange: true");
      return true;
    end

    --DebugWrite("MouseOutOfRange: false");
    return false;
  end

  self.OnLeave = function(s,a)
    --DebugWrite("self.OnLeave");

    if self:MouseOutOfRange() then
      self.Toolbar:UpdateLayout();
      self.HoverActive = false;
      self:SetBackColor(Turbine.UI.Color( self._settings.ChatOpacity, self._settings.WindowBackColor.R, self._settings.WindowBackColor.G, self._settings.WindowBackColor.B ));
      self.ToolbarBackground:SetBackColor(Turbine.UI.Color( self._settings.ToolbarOpacity, self._settings.ToolbarBackColor.R, self._settings.ToolbarBackColor.G, self._settings.ToolbarBackColor.B ));
    end

  end

  self.GetBaseToolbarButtonTop = function()
    if self._chatGroupCount > 1 and self._settings.ToolbarBottom and self._settings.Maximized then
      return self._groupHeight;
    else
      return 0;
    end
  end

  self.GetHeaderHeightWithChatGroups = function()
    if self._chatGroupCount > 1 then
      return self._headerHeight + self._groupHeight;
    else
      return self._headerHeight;
    end
  end


------------------------------------------------- CHAT WRAPPER AND WINDOWS -------------------------------------------------------------

  self.ChatWrapper = Turbine.UI.Control();
  self.ChatWrapper:SetParent(self);
  self.ChatWrapper:SetPosition(0, 0);
  self.ChatWrapper:SetHeight(self._settings.Height);
  self.ChatWrapper:SetWidth(self._settings.Width);
  self.ChatWrapper:SetMouseVisible(not self._settings.ChatClickThrough);

  self.ChatWrapper.SetActiveChatGroup = function(s,a)
    --DebugWrite("SetActiveChatGroup: " .. tostring(a));
    self._settings.ActiveChatGroupId = a;

    self.ChatGroups:SetActiveChatGroup();
    self.ChatWrapper:UpdateClearButtonColor();

    local chatWrapperWindows = self.ChatWrapper:GetControls();
    for i=1,chatWrapperWindows:GetCount() do
      if chatWrapperWindows:Get(i).Type == "Wrapper" then
        if chatWrapperWindows:Get(i).ChatGroupId == a then
          chatWrapperWindows:Get(i):SetVisible(true);
        else
          chatWrapperWindows:Get(i):SetVisible(false);
        end

        local chatWindows = chatWrapperWindows:Get(i):GetControls();
        for ii=1,chatWindows:GetCount() do
          if chatWindows:Get(ii).Type == "Chat" then
            if chatWindows:Get(ii).ChatGroupId == a then
              chatWindows:Get(ii):SetVisible(not self._settings.InteractiveMode);
              chatWindows:Get(ii):GetVerticalScrollBar():SetVisible(not self._settings.InteractiveMode);
            else
              chatWindows:Get(ii):SetVisible(false);
              chatWindows:Get(ii):GetVerticalScrollBar():SetVisible(false);
            end
          elseif chatWindows:Get(ii).Type == "InteractiveChat" then
            if chatWindows:Get(ii).ChatGroupId == a then
              chatWindows:Get(ii):SetVisible(self._settings.InteractiveMode);
              chatWindows:Get(ii):GetVerticalScrollBar():SetVisible(self._settings.InteractiveMode);
            else
              chatWindows:Get(ii):SetVisible(false);
              chatWindows:Get(ii):GetVerticalScrollBar():SetVisible(false);
            end
          end
        end
      end
    end
  end

  self.ChatWrapper.GetChatWindowsById = function(s,a)
    local chatWindow = nil;
    local interactiveChatWindow = nil;
    local chatWrapperWindows = self.ChatWrapper:GetControls();
    for i=1,chatWrapperWindows:GetCount() do
      if chatWrapperWindows:Get(i).Type == "Wrapper" then
        local chatWindows = chatWrapperWindows:Get(i):GetControls();
        for ii=1,chatWindows:GetCount() do
          if chatWindows:Get(ii).ChatGroupId == a then
            if chatWindows:Get(ii).Type == "Chat" then
              chatWindow = chatWindows:Get(ii);
            elseif chatWindows:Get(ii).Type == "InteractiveChat" then
              interactiveChatWindow = chatWindows:Get(ii);
            end
          end
        end
      end
    end
    return chatWindow, interactiveChatWindow;
  end

  self.ChatWrapper.GetInteractiveMode = function()
    local chatWindow, interactiveChatWindow = self.ChatWrapper:GetChatWindowsById(self._settings.ActiveChatGroupId);
    return interactiveChatWindow:IsVisible();
  end

  self.ChatWrapper.SetInteractiveMode = function(s,a)
    local chatWrapperWindows = self.ChatWrapper:GetControls();
    for i=1,chatWrapperWindows:GetCount() do
      if chatWrapperWindows:Get(i).Type == "Wrapper" then
        if chatWrapperWindows:Get(i).ChatGroupId == self._settings.ActiveChatGroupId then
          local chatWindows = chatWrapperWindows:Get(i):GetControls();
          for ii=1,chatWindows:GetCount() do
            if chatWindows:Get(ii).Type == "Chat" then
              chatWindows:Get(ii):SetVisible(not a);
            elseif chatWindows:Get(ii).Type == "InteractiveChat" then
              chatWindows:Get(ii):SetVisible(a);
            end
          end
        end
      end
    end
  end

  self.ChatWrapper.SetChatWindowsMouseVisible = function(s,a)
    self.ChatWrapper:SetMouseVisible(not self._settings.ChatClickThrough);
    local chatWrapperWindows = self.ChatWrapper:GetControls();
    for i=1,chatWrapperWindows:GetCount() do
      if chatWrapperWindows:Get(i).Type == "Wrapper" then
        chatWrapperWindows:Get(i):SetMouseVisible(a);
        local chatWindows = chatWrapperWindows:Get(i):GetControls();
        for ii=1,chatWindows:GetCount() do
          if chatWindows:Get(ii).Type == "Chat" or chatWindows:Get(ii).Type == "InteractiveChat" then
            chatWindows:Get(ii):SetMouseVisible(a);
            chatWindows:Get(ii):GetVerticalScrollBar():SetMouseVisible(a);
          end
        end
      end
    end
  end

  self.ChatWrapper.GetChatWindowsLeft = function(s,a)
    return self.ChatWrapper:GetLeft();
  end

  self.ChatWrapper.SetChatWindowsWidth = function(s,a)
    self.ChatWrapper:SetWidth(a);
    local chatWrapperWindows = self.ChatWrapper:GetControls();
    for i=1,chatWrapperWindows:GetCount() do
      if chatWrapperWindows:Get(i).Type == "Wrapper" then
        chatWrapperWindows:Get(i):SetWidth(a);
        local chatWindows = chatWrapperWindows:Get(i):GetControls();
        for ii=1,chatWindows:GetCount() do
          if chatWindows:Get(ii).Type == "Chat" or chatWindows:Get(ii).Type == "InteractiveChat" then
            chatWindows:Get(ii):SetWidth(self.ChatWrapper:GetWidth() - (self._scrollbarWidth + 4));
            chatWindows:Get(ii):GetVerticalScrollBar():SetLeft((self.ChatWrapper:GetLeft() + self.ChatWrapper:GetWidth()) - self._scrollbarWidth);
          end
        end
      end
    end
  end

  self.ChatWrapper.GetChatWindowsHeight = function(s,a)
    return self.ChatWrapper:GetHeight();
  end

  self.ChatWrapper.SetChatWindowsHeight = function(s,a)
    self.ChatWrapper:SetHeight(a);
    local chatWrapperWindows = self.ChatWrapper:GetControls();
    for i=1,chatWrapperWindows:GetCount() do
      if chatWrapperWindows:Get(i).Type == "Wrapper" then
        chatWrapperWindows:Get(i):SetHeight(a);
        local chatWindows = chatWrapperWindows:Get(i):GetControls();
        for ii=1,chatWindows:GetCount() do
          if chatWindows:Get(ii).Type == "Chat" or chatWindows:Get(ii).Type == "InteractiveChat" then
            chatWindows:Get(ii):SetHeight(a-4);
            chatWindows:Get(ii):GetVerticalScrollBar():SetTop(2);
            chatWindows:Get(ii):GetVerticalScrollBar():SetHeight(a-4);
          end
        end
      end
    end
  end

  self.ChatWrapper.GetChatWindowsTop = function(s,a)
    return self.ChatWrapper:GetTop();
  end

  self.ChatWrapper.SetChatWindowsTop = function(s,a)
    self.ChatWrapper:SetTop(a);
  end

  self.ChatWrapper.SetChatWindowsVisible = function(s,a)
    local chatWrapperWindows = self.ChatWrapper:GetControls();
    for i=1,chatWrapperWindows:GetCount() do
      if chatWrapperWindows:Get(i).Type == "Wrapper" and chatWrapperWindows:Get(i).ChatGroupId == self._settings.ActiveChatGroupId then
        chatWrapperWindows:Get(i):SetVisible(a);
        --[[local chatWindows = chatWrapperWindows:Get(i):GetControls();
        for ii=1,chatWindows:GetCount() do
          if chatWindows:Get(ii).Type == "Chat" or chatWindows:Get(ii).Type == "InteractiveChat" then
            chatWindows:Get(ii):SetVisible(a);
            chatWindows:Get(ii):GetVerticalScrollBar():SetVisible(a);
          end
        end]]
      end
    end

  end

  self.ChatWrapper.SetChatWindowsSize = function(s,w,h)
    self.ChatWrapper:SetSize(w,h);
    local c, i = self.ChatWrapper:GetChatWindowsById(self._settings.ActiveChatGroupId);
    c:GetParent():SetSize(w,h);
    c:SetLeft(2);
    c:SetSize((w-4) - - self._scrollbarWidth,h-4);
    i:SetLeft(2);
    i:SetSize((w-4) - - self._scrollbarWidth,h-4);
    c:GetVerticalScrollBar():SetTop(2);
    c:GetVerticalScrollBar():SetHeight(self.ChatWrapper:GetHeight()-4);
    c:GetVerticalScrollBar():SetLeft((self.ChatWrapper:GetLeft() + self.ChatWrapper:GetWidth()) - self._scrollbarWidth);
  end

  self.ChatWrapper.SetChatWindowsForeColor = function(s,a)
    local chatWrapperWindows = self.ChatWrapper:GetControls();
    for i=1,chatWrapperWindows:GetCount() do
      if chatWrapperWindows:Get(i).Type == "Wrapper" then
        local chatWindows = chatWrapperWindows:Get(i):GetControls();
        for ii=1,chatWindows:GetCount() do
          if chatWindows:Get(ii).Type == "Chat" or chatWindows:Get(ii).Type == "InteractiveChat" then
            chatWindows:Get(ii):SetForeColor(a);
          end
        end
      end
    end
  end

  ---@diagnostic disable-next-line: duplicate-set-field
  self.ChatWrapper.MouseHover = function(sender,args)
    self:OnHover();
  end

  ---@diagnostic disable-next-line: duplicate-set-field
  self.ChatWrapper.MouseLeave = function(sender,args)
    self:OnLeave();
  end

  self.ChatWrapper.UpdateClearButtonColor = function(s)
    local textLength = self:GetTextLength(self._settings.ActiveChatGroupId);
    if textLength > 60000 then
      -- Max label length is 64k, so force remove ~32k from the beginning of the text
      self.Clear:SetForeColor(Turbine.UI.Color(1, self._settings.WarningColor.R, self._settings.WarningColor.G, self._settings.WarningColor.B));
      self:SetText(self:GetText():sub(-32000));

      -- Remove ~28k from chat message log
      local strLen = 0;
      for key,value in spairs(self._entityMessages, function(t,aa,bb) return t[bb].LocalTime > t[aa].LocalTime end) do
        strLen = strLen + string.len(self._entityMessages[key].Message);
        self._entityMessages[key] = nil;
        if strLen > 28000 then
          break;
        end
      end
    elseif textLength > 32000 then
      self.Clear:SetForeColor(Turbine.UI.Color(1, self._settings.WarningColor.R, self._settings.WarningColor.G, self._settings.WarningColor.B));
    elseif textLength > 16000 then
      self.Clear:SetForeColor(Turbine.UI.Color(1, self._settings.ToggleColor.R, self._settings.ToggleColor.G, self._settings.ToggleColor.B));
    end
  end

  self.ChatWrapper.UpdateChatGroupControls = function(s)

    local chatControls = self.ChatWrapper:GetControls();
    for i=chatControls:GetCount(),1,-1 do
      local deleteGroup = true;
      for key,value in pairs(self._settings.ChatGroups) do
        if chatControls.ChatGroupId == value.Id then
          deleteGroup = false;
        end
      end

      if deleteGroup then
        chatControls:RemoveAt(i);
      end
    end

    for key,value in pairs(self._settings.ChatGroups) do
      --DebugWrite("self._settings.ChatGroups: key - " .. tostring(key));
      --DebugWrite("self._settings.ChatGroups: value - " .. tostring(value));
      --DebugWrite("self._settings.ChatGroups: value.Id - " .. tostring(value.Id));
      -- Chat label with custom tags stripped out

      local addGroup = true;
      chatControls = self.ChatWrapper:GetControls();
      for k,v in pairs(chatControls) do
        if chatControls.ChatGroupId == value.Id then
          addGroup = false;
        end
      end

      if addGroup then

        local groupWrapper = Turbine.UI.Control();
        groupWrapper.Type = "Wrapper";
        groupWrapper.ChatGroupId = value.Id;
        groupWrapper:SetParent(self.ChatWrapper);
        groupWrapper:SetSize(self._settings.Width,self._settings.Height);
        groupWrapper:SetPosition(0,0);
        groupWrapper:SetMouseVisible(not self._settings.ChatClickThrough);

        -- Both chat labels share the same scrollbar since they display the same text length
        local scrollBar = Turbine.UI.Lotro.ScrollBar();
        scrollBar.ChatGroupId = value.Id;
        scrollBar.Type = "Scrollbar";
        scrollBar:SetOrientation(Turbine.UI.Orientation.Vertical);
        scrollBar:SetParent(groupWrapper);
        scrollBar:SetPosition((self.ChatWrapper:GetLeft() + self.ChatWrapper:GetWidth()) - self._scrollbarWidth, 2);
        scrollBar:SetHeight(self.ChatWrapper:GetHeight() - 4);
        scrollBar:SetWidth(self._scrollbarWidth);
        --scrollBar:SetVisible(value.Id == self._settings.ActiveChatGroupId);

        local chatText = Turbine.UI.Label();
        chatText.Type = "Chat";
        chatText.ChatGroupId = value.Id;
        chatText:SetParent(groupWrapper);
        chatText:SetFont(Turbine.UI.Lotro.Font.Verdana16);
        chatText:SetFontStyle(Turbine.UI.FontStyle.Outline);
        chatText:SetMarkupEnabled(true);
        chatText:SetMultiline( true );
        chatText:SetSelectable( true );
        chatText:SetForeColor( Turbine.UI.Color(1, self._settings.WindowForeColor.R, self._settings.WindowForeColor.B, self._settings.WindowForeColor.G) );
        chatText:SetPosition(2, 2);
        chatText:SetSize(self.ChatWrapper:GetWidth() - (chatText:GetLeft() + self._scrollbarWidth + 2), self.ChatWrapper:GetHeight() - 4);
        chatText:SetMouseVisible(not self._settings.ChatClickThrough);
        chatText:SetVerticalScrollBar(scrollBar);

        ---@diagnostic disable-next-line: duplicate-set-field
        chatText.MouseHover = function(sender,args)
          self:OnHover();
        end

        ---@diagnostic disable-next-line: duplicate-set-field
        chatText.MouseLeave = function(sender,args)
          self:OnLeave();
        end

        --chatText:SetVisible(false);
        --if self._settings.Maximized then
          if chatText.ChatGroupId == self._settings.ActiveChatGroupId then
            chatText:SetVisible(not self._settings.InteractiveMode);
          end
        --end


        -- Chat label with clickable links
        local interactiveChatText = Turbine.UI.Label();
        interactiveChatText.Type = "InteractiveChat";
        interactiveChatText.ChatGroupId = value.Id;
        interactiveChatText:SetParent(groupWrapper);
        interactiveChatText:SetFont(Turbine.UI.Lotro.Font.Verdana16);
        interactiveChatText:SetFontStyle(Turbine.UI.FontStyle.Outline);
        interactiveChatText:SetMarkupEnabled(true);
        interactiveChatText:SetMultiline( true );
        interactiveChatText:SetSelectable( true );
        interactiveChatText:SetForeColor( Turbine.UI.Color( 0, 0.7, 0 ) );
        interactiveChatText:SetPosition(2,2);
        interactiveChatText:SetSize(self.ChatWrapper:GetWidth() - (chatText:GetLeft() + self._scrollbarWidth + 2), self.ChatWrapper:GetHeight() - 4);
        interactiveChatText:SetMouseVisible(not self._settings.ChatClickThrough);
        interactiveChatText:SetVerticalScrollBar(scrollBar);

        ---@diagnostic disable-next-line: duplicate-set-field
        interactiveChatText.MouseHover = function(sender,args)
          self:OnHover();
        end

        ---@diagnostic disable-next-line: duplicate-set-field
        interactiveChatText.MouseLeave = function(sender,args)
          self:OnLeave();
        end

        --interactiveChatText:SetVisible(false);
        --if self._settings.Maximized then
          if interactiveChatText.ChatGroupId == self._settings.ActiveChatGroupId then
            interactiveChatText:SetVisible(self._settings.InteractiveMode);
          end
        --end

        if interactiveChatText.ChatGroupId == self._settings.ActiveChatGroupId then
          interactiveChatText:GetVerticalScrollBar():SetVisible(true);
        end

        chatText:SetText(" ");
        chatText:SetText("");
        interactiveChatText:SetText(" ");
        interactiveChatText:SetText("");

        groupWrapper:SetVisible(self._settings.Maximized);
      end
    end
  end

  self.ChatWrapper:UpdateChatGroupControls();

  ------------------------------------------------- TOOLBAR BACKGROUND -------------------------------------------------------------


  self.ToolbarBackground = Turbine.UI.Control();
  self.ToolbarBackground:SetParent (self);
  if self._settings.ToolbarBottom then
    self.ToolbarBackground:SetPosition(0, self:GetHeight() - self:GetHeaderHeightWithChatGroups());
  else
    self.ToolbarBackground:SetPosition(0, 0);
  end
  self.ToolbarBackground:SetSize(self:GetWidth(), self:GetHeaderHeightWithChatGroups());
  self.ToolbarBackground:SetMouseVisible(false);
  self.ToolbarBackground:SetOpacity(self._settings.ChatOpacity)
  self.ToolbarBackground:SetBackColor(Turbine.UI.Color(self._settings.ToolbarOpacity, self._settings.ToolbarBackColor.R, self._settings.ToolbarBackColor.G, self._settings.ToolbarBackColor.B));
  self.ToolbarBackground:SetZOrder(-1);
  self.ToolbarBackground:SetVisible(self._settings.ToolbarPinned);


  ------------------------------------------ TOOLBAR - NAME AND BUTTONS CONTAINER --------------------------------------------------------


  self.Toolbar = Turbine.UI.Control();
  self.Toolbar:SetParent (self);
  self.Toolbar:SetSize(self:GetWidth(), self:GetHeaderHeightWithChatGroups());
  self.Toolbar:SetMouseVisible(true);
  self.Toolbar:SetOpacity(self._settings.ChatOpacity)
  self.Toolbar:SetZOrder(0)
  self.Toolbar.RightButtonDown = false;
  self.Toolbar.IsBeingDragged = false;
  self.Toolbar.TipText = "Control + left mouse to move window";

  if self._settings.ToolbarBottom then
    self.Toolbar:SetPosition(0, self:GetHeight() - self._headerHeight);
  else
    self.Toolbar:SetPosition(0, 0);
  end

  if self._settings.Maximized then
    if not self._settings.ToolbarPinned then
      self.ToolbarBackground:SetHeight(5);
      self.Toolbar:SetHeight(5);
    end
  else
    self.ToolbarBackground:SetHeight(self:GetHeaderHeightWithChatGroups());
    self.Toolbar:SetHeight(self:GetHeaderHeightWithChatGroups());
  end

  self.Toolbar.UpdateLayout = function(sender,args)

    --DebugWrite("self.Toolbar.UpdateLayout");

    self.ChatGroups:UpdateGroups();

    if self._settings.Maximized then
      if args == nil then
        if not self._settings.ToolbarPinned then
          --DebugWrite("A")
          self.ToolbarBackground:SetHeight(5);
          self.Toolbar:SetHeight(5);
        end
      end

      if self._settings.ToolbarBottom then
        if self.ChatGroups:GetItemCount() > 1 then
          self.Toolbar:SetTop(self:GetHeight() - (self._headerHeight + self._groupHeight));
          self.Toolbar:SetHeight(self._headerHeight + self._groupHeight);
        else
          self.Toolbar:SetTop(self:GetHeight() - self._headerHeight);
          self.Toolbar:SetHeight(self._headerHeight);
        end
      else
        self.Toolbar:SetTop(0);
        if self.ChatGroups:GetItemCount() > 1 then
          self.Toolbar:SetHeight(self._headerHeight + self._groupHeight);
        else
          self.Toolbar:SetHeight(self._headerHeight);
        end
        --self.Toolbar:ShowButtons(false);
      end

    else
      --DebugWrite("B")

      self.ChatGroups:SetTop(self._groupHeight);
      self.ToolbarBackground:SetHeight(self._headerHeight);
      self.Toolbar:SetTop(0);
      self.Toolbar:SetHeight(self._headerHeight);

    end

    self.ToolbarBackground:SetTop(self.Toolbar:GetTop());
    self.ToolbarBackground:SetHeight(self.Toolbar:GetHeight());
    self.ChatGroups:UpdateLayout();

    if args == nil then
      if not self._settings.ToolbarPinned then
        self.ToolbarBackground:SetVisible(false);
        self.Toolbar:ShowButtons(false);
      end
    end

    self.ChatWrapper:SetChatWindowsTop(0);
    if not self._settings.ToolbarBottom and self._settings.ToolbarPinned then
      self.ChatWrapper:SetChatWindowsTop(self:GetHeaderHeightWithChatGroups() + 0);
    end

    if self._settings.ToolbarPinned then
      self.ChatWrapper:SetChatWindowsHeight(self._settings.Height - (self:GetHeaderHeightWithChatGroups() + 0));
    else
      self.ChatWrapper:SetChatWindowsHeight(self._settings.Height - 0);
    end

    self.Name:SetWidth(self._settings.Width - self._leftButtonPos - (self._settings.Width - self._rightButtonPos));

    local buttons = self.Toolbar:GetControls();
    for i=1,buttons:GetCount() do
      if buttons:Get(i).UpdateButton ~= nil then
        buttons:Get(i):UpdateButton();
      end
    end
    self.Name:UpdateVisibility();

  end

  self.Toolbar.ShowButtons = function(sender, value)
    local buttons = self.Toolbar:GetControls();
    for i=1,buttons:GetCount() do
      buttons:Get(i):SetVisible(value);
    end

    self.ChatGroups:SetVisible(self.ChatGroups:GetItemCount() > 1 and value and self._settings.Maximized);
    for i=1,self.ChatGroups:GetItemCount() do
      self.ChatGroups:GetItem(i):SetVisible(self.ChatGroups:GetItemCount() > 1 and value);
    end
    self.MinMax:UpdateButton();

    if false then
      DebugWrite("self.Toolbar.ShowButtons");
      DebugWrite("  --- self.Toolbar.GetTop(): "..tostring(self.Toolbar:GetTop()));
      DebugWrite("  --- self.Toolbar.GetLeft(): "..tostring(self.Toolbar:GetLeft()));
      DebugWrite("  --- self.Toolbar.GetHeight(): "..tostring(self.Toolbar:GetHeight()));
      DebugWrite("  --- self.MinMax.GetTop(): "..tostring(self.MinMax:GetTop()));
      DebugWrite("  --- self.MinMax.GetLeft(): "..tostring(self.MinMax:GetLeft()));
      DebugWrite("  --- self.MinMax.GetHeight(): "..tostring(self.MinMax:GetHeight()));
      DebugWrite("  --- self.MinMax.IsVisible(): "..tostring(self.MinMax:IsVisible()));
      DebugWrite("  --- self.MinMax2.GetTop(): "..tostring(self.MinMax2:GetTop()));
      DebugWrite("  --- self.MinMax2.GetLeft(): "..tostring(self.MinMax2:GetLeft()));
      DebugWrite("  --- self.MinMax2.GetHeight(): "..tostring(self.MinMax2:GetHeight()));
      DebugWrite("  --- self.MinMax2.IsVisible(): "..tostring(self.MinMax2:IsVisible()));
    end
  end

  ---@diagnostic disable-next-line: duplicate-set-field
  self.Toolbar.MouseHover = function(sender,args)
    --DebugWrite("self.Toolbar.MouseHover");
    self:OnHover(sender);
    if self._settings.TipsEnabled then
      Libraries.Hint( sender, sender.TipText)
    end
  end

  ---@diagnostic disable-next-line: duplicate-set-field
  self.Toolbar.MouseEnter = function(sender,args)
    --DebugWrite("self.Toolbar.MouseEnter");
    if not self._settings.ToolbarPinned then

      if self._settings.ToolbarBottom then
        self.Toolbar:SetTop(self:GetHeight() - self:GetHeaderHeightWithChatGroups());
      end

      self.Toolbar:SetHeight(self:GetHeaderHeightWithChatGroups());
      self.ToolbarBackground:SetTop(self.Toolbar:GetTop());
      self.ToolbarBackground:SetHeight(self:GetHeaderHeightWithChatGroups());
      self.ToolbarBackground:SetVisible(true);

      if not self._settings.ToolbarBottom then
        self.ChatWrapper:SetChatWindowsTop(self:GetHeaderHeightWithChatGroups());
      end

      self.ChatWrapper:SetChatWindowsHeight(self._settings.Height-(self:GetHeaderHeightWithChatGroups() + 0));

      self.Toolbar:ShowButtons(true);
    elseif not self._settings.Maximized then
      self.Toolbar:ShowButtons(true);
    end

  end

  ---@diagnostic disable-next-line: duplicate-set-field
  self.Toolbar.MouseLeave = function(sender,args)
    --DebugWrite("self.Toolbar.MouseLeave");
    Libraries.Hint:Remove();

    if self:MouseOutOfRange() then
      if not self._settings.ToolbarPinned then
          self.Toolbar:UpdateLayout();
      end
      self:OnLeave();
    end

  end

  ---@diagnostic disable-next-line: duplicate-set-field
  self.Toolbar.MouseDown = function(sender,args)
    --DebugWrite("self.Toolbar.MouseDown");
    Libraries.Hint:Remove();
    if args.Button == Turbine.UI.MouseButton.Left and self.Toolbar:IsControlKeyDown() then
      self.Toolbar.IsBeingDragged = true;
      self._dragStartX = args.X;
      self._dragStartY = args.Y;
    elseif args.Button == Turbine.UI.MouseButton.Right then
      self.Toolbar.RightButtonDown = true;
    end

  end

  ---@diagnostic disable-next-line: duplicate-set-field
  self.Toolbar.MouseMove = function( sender, args )

		if self.Toolbar.IsBeingDragged then
			local left, top = self:GetPosition();
			self:SetPosition( math.min( math.max( left + ( args.X - self._dragStartX ), 0 ), Turbine.UI.Display:GetWidth() - self:GetWidth() )
									  , math.min( math.max( top + ( args.Y - self._dragStartY ), 0 ), Turbine.UI.Display:GetHeight() - self:GetHeight()) );
    end

  end

  ---@diagnostic disable-next-line: duplicate-set-field
  self.Toolbar.MouseUp = function(sender,args)
    --DebugWrite("self.Toolbar.MouseUp");
    if args.Button == Turbine.UI.MouseButton.Left and self.Toolbar.IsBeingDragged then
      self.Toolbar.IsBeingDragged = false;
      self._dragStartX = 0;
      self._dragStartY = 0;

      self._settings.Left = self:GetLeft();
      self._settings.Top = self:GetTop();
    elseif args.Button == Turbine.UI.MouseButton.Right then
      self.Toolbar.RightButtonDown = false;
    end

  end

  ---@diagnostic disable-next-line: duplicate-set-field
  self.Toolbar.MouseClick = function(sender,args)
    --DebugWrite("self.Toolbar.MouseClick");
    if args.Button == Turbine.UI.MouseButton.Right then

      self.Toolbar.UpdateLayout();

      local contextMenu = Turbine.UI.ContextMenu();

      ---------------------------------------------------------------------------
      -- SETTINGS MENU ITEM AND WINDOW ------------------------------------------
      ---------------------------------------------------------------------------
      local menuSettings = Turbine.UI.MenuItem("Settings", true);
      ---@diagnostic disable-next-line: duplicate-set-field
      menuSettings.Click = function(s,a)

        ChatWindowSettingsWindow = Turbine.UI.Lotro.GoldWindow();
        ChatWindowSettingsWindow.ChatWindow = self;
        ChatWindowSettingsWindow:SetHeight(720);
        ChatWindowSettingsWindow:SetWidth(800);
        ChatWindowSettingsWindow:SetLeft((Turbine.UI.Display:GetWidth() / 2) - (ChatWindowSettingsWindow:GetWidth() / 2));
        ChatWindowSettingsWindow:SetTop(((Turbine.UI.Display:GetHeight() / 2) - (ChatWindowSettingsWindow:GetHeight() / 2)) - 75);

        ChatWindowSettingsWindow.LoadGroups = function()

          local groupsHelpHeader = Turbine.UI.Label();
          groupsHelpHeader:SetParent(ChatWindowSettingsWindow);
          groupsHelpHeader:SetFont(Turbine.UI.Lotro.Font.TrajanPro14);
          groupsHelpHeader:SetMultiline(true);
          groupsHelpHeader:SetTextAlignment(Turbine.UI.ContentAlignment.TopLeft);
          groupsHelpHeader:SetSize(ChatWindowSettingsWindow.ChannelGroupsListBox:GetWidth() , 30);
          groupsHelpHeader:SetTop(ChatWindowSettingsWindow.ChannelGroupsListBox:GetTop() - 30);
          groupsHelpHeader:SetLeft(ChatWindowSettingsWindow.ChannelGroupsListBox:GetLeft());
          groupsHelpHeader:SetText(" Chat Groups\n Change order ( ^ or v ) or delete ( X ).");

          local spacerG = Turbine.UI.Label();
          spacerG:SetParent(ChatWindowSettingsWindow);
          spacerG:SetSize(ChatWindowSettingsWindow.ChannelGroupsListBox:GetWidth(), 1);
          spacerG:SetBackColor(Turbine.UI.Color.Gray);
          spacerG:SetTop(groupsHelpHeader:GetTop() + groupsHelpHeader:GetHeight());
          spacerG:SetLeft(ChatWindowSettingsWindow.ChannelGroupsListBox:GetLeft());

          local addWrapper = Turbine.UI.Control();
          addWrapper:SetSize(ChatWindowSettingsWindow.ChannelGroupsListBox:GetWidth(), 28);
          addWrapper:SetParent(ChatWindowSettingsWindow);
          addWrapper:SetTop(ChatWindowSettingsWindow.ChannelGroupsListBox:GetTop() + ChatWindowSettingsWindow.ChannelGroupsListBox:GetHeight() + 5)
          addWrapper:SetLeft(ChatWindowSettingsWindow.ChannelGroupsListBox:GetLeft());

          local newLabel = Turbine.UI.Label();
          newLabel:SetParent(addWrapper);
          newLabel:SetFont(Turbine.UI.Lotro.Font.TrajanPro18);
          newLabel:SetText("New Group:");
          newLabel:SetLeft(0);
          newLabel:SetTop(8);
          newLabel:SetWidth(100);
          newLabel:SetVisible(true);

          local newTextBox = Turbine.UI.Lotro.TextBox();
          newTextBox:SetFont(Turbine.UI.Lotro.Font.TrajanPro18);
          newTextBox:SetParent(addWrapper);
          newTextBox:SetLeft(newLabel:GetLeft() + newLabel:GetWidth() + 2);
          newTextBox:SetTop(6);
          newTextBox:SetWidth(140);
          newTextBox:SetHeight(20);
          newTextBox:SetVisible(true);
          newTextBox:SetMultiline(false);

          local newButton = Turbine.UI.Lotro.Button();
          newButton:SetParent(addWrapper);
          newButton:SetLeft(newTextBox:GetLeft() + newTextBox:GetWidth() + 6);
          newButton:SetFont(Turbine.UI.Lotro.Font.TrajanPro16)
          newButton:SetForeColor(Turbine.UI.Color.White);
          newButton:SetTop(6);
          newButton:SetSize(50, 28);
          newButton:SetText("Add");
          newButton.TextBox = newTextBox;
          ---@diagnostic disable-next-line: duplicate-set-field
          newButton.MouseClick = function(sc,sa)
            local txt = sc.TextBox:GetText();
            if txt == nil or txt == "" or string.len(string.gsub(txt, "%s+", "")) == 0 then
              return;
            else
              table.insert(self._settings.ChatGroups,
                           {
                            Id = self._chatGroupMaxId + 1,
                            Order = self._chatGroupMaxOrder + 1,
                            Name = txt,
                            Channels = {}
                           }
              );
              self._settings.Enabled = false;
              self:UpdateChatGroups();
              Save();
              self._settings.Enabled = true;
              self.ChatWrapper:UpdateChatGroupControls();
              self.ChatGroups:UpdateGroups();
              self.Toolbar:MouseLeave({});
              ChatWindowSettingsWindow:LoadGroups();
            end
          end

          local spacer = Turbine.UI.Label();
          spacer:SetParent(addWrapper);
          spacer:SetSize(ChatWindowSettingsWindow.ChannelGroupsListBox:GetWidth(), 1);
          spacer:SetBackColor(Turbine.UI.Color.Gray);
          spacer:SetTop(0);

          ChatWindowSettingsWindow.ChannelGroupsListBox:ClearItems();

          local loopct = 1;
          for key,value in spairs(self._settings.ChatGroups, function(t,aa,bb) return t[bb].Order > t[aa].Order end) do
            local wrapper = Turbine.UI.Control();
            wrapper:SetSize(ChatWindowSettingsWindow.ChannelGroupsListBox:GetWidth(), 50);
            wrapper.ChatGroupId = value.Id;

            if loopct > 1 then
              local spacerI = Turbine.UI.Label();
              spacerI:SetParent(wrapper);
              spacerI:SetSize(ChatWindowSettingsWindow.ChannelGroupsListBox:GetWidth(), 1);
              spacerI:SetLeft(0);
              spacerI:SetBackColor(Turbine.UI.Color.Gray);
              spacerI:SetTop(0);
            end

            local groupNameTextBox = Turbine.UI.Lotro.TextBox();
            groupNameTextBox:SetParent(wrapper);
            groupNameTextBox:SetFont(Turbine.UI.Lotro.Font.TrajanPro18);
            groupNameTextBox:SetTop(16);
            groupNameTextBox:SetLeft(54);
            groupNameTextBox:SetSize(140, 20);
            groupNameTextBox:SetMultiline(false);
            groupNameTextBox:SetText(value.Name);
            groupNameTextBox.OriginalName = value.Name;
            groupNameTextBox.ChatGroupId = value.Id;
            ---@diagnostic disable-next-line: duplicate-set-field
            groupNameTextBox.FocusLost = function(sc,sa)
              local txt = sc:GetText();
              if txt == nil or txt == "" or string.len(string.gsub(txt, "%s+", "")) == 0 then
                sc:SetText(sc.OriginalName);
                return;
              end
              if sc.OriginalName ~= txt then
                sc.OriginalName = txt;
                self._settings.ChatGroups[sc.ChatGroupId].Name = txt;
                self:UpdateChatGroups();
                Save();
                self.ChatWrapper:UpdateChatGroupControls();
                self.ChatGroups:UpdateGroups();
                self.Toolbar:MouseLeave({});
              end
            end

            local upButton = Turbine.UI.Lotro.Button();
            upButton:SetParent(wrapper);
            upButton:SetLeft(ChatWindowSettingsWindow.ChannelGroupsListBox:GetWidth() - 48);
            upButton:SetFont(Turbine.UI.Lotro.Font.TrajanPro21)
            upButton:SetForeColor(iif(self._chatGroupCount > 1 and loopct > 1, Turbine.UI.Color.White, Turbine.UI.Color.DarkGray));
            upButton:SetTop(4);
            upButton:SetSize(20, 20);
            upButton:SetText("^");
            upButton:SetEnabled(self._chatGroupCount > 1 and loopct > 1);
            upButton.ChatGroupId = value.Id;

            ---@diagnostic disable-next-line: duplicate-set-field
            upButton.MouseClick = function(sc,sa)
              --DebugWrite("upButton.MouseClick");
              for k,v in spairs(self._settings.ChatGroups, function(t,aa,bb) return t[bb].Order < t[aa].Order end) do
                --DebugWrite("Chat Group Loop Order: ID: " .. tostring(v.Id) .. " - Order: " .. tostring(v.Order));
                --DebugWrite("   button object data: ID: " .. tostring(sc.ChatGroupId) .. " - Order: " .. tostring(sc.ChatGroupOrder));
                if sc.ChatGroupId == v.Id then
                  --DebugWrite("Chat Group ID Match - setting current Order to " .. tostring(v.Order - 1));
                  v.Order = v.Order - 1;
                  sc.ChatGroupOrder = v.Order;
                elseif sc.ChatGroupId ~= v.Id and sc.ChatGroupOrder == v.Order then
                  --DebugWrite("Chat Order Match - setting next Order to " .. tostring(v.Order + 1));
                  v.Order = v.Order + 1;
                  sc.ChatGroupOrder = v.Order;
                end
              end
              self:UpdateChatGroups();
              Save();
              self.ChatWrapper:UpdateChatGroupControls();
              self.ChatGroups:UpdateGroups();
              self.Toolbar:MouseLeave({});
              ChatWindowSettingsWindow:LoadGroups();
            end

            local downButton = Turbine.UI.Lotro.Button();
            downButton:SetParent(wrapper);
            downButton:SetFont(Turbine.UI.Lotro.Font.TrajanPro15)
            downButton:SetLeft(upButton:GetLeft());
            downButton:SetForeColor(iif(self._chatGroupCount > 1 and loopct < self._chatGroupCount , Turbine.UI.Color.White, Turbine.UI.Color.DarkGray));
            downButton:SetTop(27);
            downButton:SetSize(20, 20);
            downButton:SetText("v");
            downButton:SetEnabled(self._chatGroupCount > 1 and loopct < self._chatGroupCount );
            downButton.ChatGroupId = value.Id;
            downButton.ChatGroupOrder = value.Order;

            ---@diagnostic disable-next-line: duplicate-set-field
            downButton.MouseClick = function(sc,sa)
              --DebugWrite("downButton.MouseClick");
              for k,v in spairs(self._settings.ChatGroups, function(t,aa,bb) return t[bb].Order > t[aa].Order end) do
                --DebugWrite("Chat Group Loop Order: ID: " .. tostring(v.Id) .. " - Order: " .. tostring(v.Order));
                --DebugWrite("   button object data: ID: " .. tostring(sc.ChatGroupId) .. " - Order: " .. tostring(sc.ChatGroupOrder));
                if sc.ChatGroupId == v.Id then
                  --DebugWrite("Chat Group ID Match - setting current Order to " .. tostring(v.Order + 1));
                  v.Order = v.Order + 1;
                  sc.ChatGroupOrder = v.Order;
                elseif sc.ChatGroupId ~= v.Id and sc.ChatGroupOrder == v.Order then
                  --DebugWrite("Chat Order Match - setting next Order to " .. tostring(v.Order - 1));
                  v.Order = v.Order - 1;
                  sc.ChatGroupOrder = v.Order;
                end
              end
              self:UpdateChatGroups();
              Save();
              self.ChatWrapper:UpdateChatGroupControls();
              self.ChatGroups:UpdateGroups();
              self.Toolbar:MouseLeave({});
              ChatWindowSettingsWindow:LoadGroups();
            end

            local deleteButton = Turbine.UI.Lotro.Button();
            deleteButton:SetParent(wrapper);
            deleteButton:SetFont(Turbine.UI.Lotro.Font.TrajanPro15)
            deleteButton:SetLeft(7);
            deleteButton:SetForeColor(iif(self._chatGroupCount > 1, Turbine.UI.Color.Red, Turbine.UI.Color.DarkGray));
            deleteButton:SetTop(16);
            deleteButton:SetSize(20, 20);
            deleteButton:SetText("X");
            deleteButton:SetEnabled(self._chatGroupCount > 1);
            deleteButton.ChatGroupId = value.Id;
            deleteButton.ChatGroupName = value.Name;

            ---@diagnostic disable-next-line: duplicate-set-field
            deleteButton.MouseClick = function(sc,sa)
              local tempMsg = Libraries.MessageBox("Delete group '".. sc.ChatGroupName .. "'?","MBYESNO");
              tempMsg.ButtonClick = function (ssc,ssa) 
                if ssa.Button == "MBYES" then
                  --DebugWrite("deleteButton.MouseClick - YES");
                  local newCount = self._chatGroupCount;
                  for i=self._chatGroupCount,1,-1 do
                    --DebugWrite("Chat Group Loop Order: ID: " .. tostring(self._settings.ChatGroups[i].Id) .. " - Order: " .. tostring(self._settings.ChatGroups[i].Order));
                    --DebugWrite("   button object data: ID: " .. tostring(sc.ChatGroupId) .. " - Order: " .. tostring(sc.ChatGroupOrder));
                    if sc.ChatGroupId == self._settings.ChatGroups[i].Id then
                      --DebugWrite("Chat Group ID Match - setting current group to nil");
                      table.remove(self._settings.ChatGroups, i);
                      newCount = newCount - 1;
                    end
                  end
                  self._chatGroupCount = newCount;

                  local loopOrder = 1;
                  for k,v in spairs(self._settings.ChatGroups, function(t,aa,bb) return t[bb].Order > t[aa].Order end) do
                    v.Order = loopOrder;
                    loopOrder= loopOrder + 1;
                  end
                  self:UpdateChatGroups();
                  Save();
                  self.ChatWrapper:UpdateChatGroupControls();
                  self.ChatGroups:UpdateGroups();
                  self.Toolbar:MouseLeave({});
                  ChatWindowSettingsWindow:LoadGroups();
                end
              end
            end

            ChatWindowSettingsWindow.ChannelGroupsListBox:AddItem(wrapper);

            loopct = loopct + 1;
          end

          ChatWindowSettingsWindow.ChannelGroupsListBox:SetSelectedIndex(1);
        end

        ChatWindowSettingsWindow.LoadChatGroup = function(sc,ac)
          ChatWindowSettingsWindow.ChannelListBox:ClearItems();

          if self._settings.ChatGroups[ac] == nil then
            return;
          end

          local groupNameWrapper = Turbine.UI.Control();
          groupNameWrapper:SetSize(ChatWindowSettingsWindow.ChannelListBox:GetWidth(),32);
          groupNameWrapper:SetBackColor(Turbine.UI.Color(0.9,0.15,0.15,0.15));
          local groupNameLabel = Turbine.UI.Label();

          groupNameLabel.ChatType = -1;
          groupNameLabel:SetParent(groupNameWrapper);
          groupNameLabel:SetFont(Turbine.UI.Lotro.Font.TrajanProBold22);
          groupNameLabel:SetTop(5);
          groupNameLabel:SetLeft(6);
          groupNameLabel:SetSize(ChatWindowSettingsWindow.ChannelListBox:GetWidth(),20);
          groupNameLabel:SetVisible(true);
          groupNameLabel:SetText("'" .. self._settings.ChatGroups[ac].Name .. "' Chat Group Channels");

          local spacerI = Turbine.UI.Label();
          spacerI:SetParent(groupNameWrapper);
          spacerI:SetSize(ChatWindowSettingsWindow.ChannelListBox:GetWidth(), 1);
          spacerI:SetBackColor(Turbine.UI.Color.Gray);
          spacerI:SetTop(groupNameWrapper:GetHeight() - 1);

          ChatWindowSettingsWindow.ChannelListBox:AddItem(groupNameWrapper);

          local loopct = 1;
          for key,value in spairs(_G.ChannelNames, function(t,aa,bb) return t[bb] > t[aa] end) do

            local chatType = key;
            local isChecked = self._settings.ChatGroups[ac].Channels[chatType];

            local wrapper = Turbine.UI.Control();
            wrapper.ChatType = chatType;
            wrapper:SetSize(ChatWindowSettingsWindow.ChannelListBox:GetWidth() - 10, 26);
            wrapper:SetBackColor(Turbine.UI.Color(0.8,0,0,0));

            local wrapperBG = Turbine.UI.Control();
            wrapperBG:SetParent(wrapper);
            wrapperBG:SetSize((ChatWindowSettingsWindow.ChannelListBox:GetWidth() - 10) - 26, 26);
            wrapperBG:SetLeft(32);
            wrapperBG:SetBackColor(Turbine.UI.Color(0.9,0,0,0));

            local colorPickerLabel = Turbine.UI.Label();
            colorPickerLabel.ChatType = chatType;
            colorPickerLabel:SetParent(wrapper);
            colorPickerLabel:SetBackColor(Turbine.UI.Color( ChatWindowSettingsWindow.ChatWindow.HexToColor(ChatWindowSettingsWindow.ChatWindow._settings.ChatGroups[ac].ChannelColors[chatType], nil)));
            colorPickerLabel:SetForeColor(colorPickerLabel:GetBackColor());
            colorPickerLabel:SetTop(6);
            colorPickerLabel:SetLeft(6);
            colorPickerLabel:SetSize(20,15);
            colorPickerLabel:SetVisible(true);
            ---@diagnostic disable-next-line: duplicate-set-field
            colorPickerLabel.MouseClick = function(f,ar)
              ChatWindowSettingsWindow.ChatWindow:PickChatColor({ ChatGroupId = self._settings.ChatGroups[ac].Id, ChatType = f.ChatType, Label = f});
            end

            local checkBox = Turbine.UI.Lotro.CheckBox();
            checkBox.ChatType = chatType;
            checkBox:SetFont(Turbine.UI.Lotro.Font.TrajanPro18);
            checkBox:SetParent(wrapper);
            checkBox:SetTop(3);
            checkBox:SetLeft(38);
            checkBox:SetHeight(22);
            checkBox:SetWidth(ChatWindowSettingsWindow.ChannelListBox:GetWidth() - checkBox:GetLeft());
            checkBox:SetText(" " .. value);
            checkBox:SetChecked(isChecked);
            checkBox:SetVisible(true);
            ---@diagnostic disable-next-line: duplicate-set-field
            checkBox.CheckedChanged = function(cb)
              ChatWindowSettingsWindow.ChatWindow._settings.ChatGroups[ac].Channels[cb.ChatType] = cb:IsChecked();
              Save();
            end

            if chatType == Turbine.ChatType.SelfLoot or chatType == Turbine.ChatType.Kinship then

              checkBox:SetWidth(checkBox:GetWidth() - 200);

              local secondaryCheckBox = Turbine.UI.Lotro.CheckBox();
              secondaryCheckBox.SelfLootNoCoins = chatType == Turbine.ChatType.SelfLoot;
              secondaryCheckBox.KinNoAnnounce = chatType == Turbine.ChatType.Kinship;
              secondaryCheckBox:SetFont(Turbine.UI.Lotro.Font.TrajanPro18);
              secondaryCheckBox:SetParent(wrapper);
              secondaryCheckBox:SetLeft(checkBox:GetLeft() + checkBox:GetWidth());
              secondaryCheckBox:SetTop(checkBox:GetTop());
              secondaryCheckBox:SetHeight(22);
              secondaryCheckBox:SetWidth(150);
              if secondaryCheckBox.SelfLootNoCoins then
                secondaryCheckBox:SetText(" " .. "No Coins");
                secondaryCheckBox:SetChecked(ChatWindowSettingsWindow.ChatWindow._settings.ChatGroups[ac].SelfLootNoCoins);
              elseif secondaryCheckBox.KinNoAnnounce then
                secondaryCheckBox:SetText(" " .. "No Announce");
                secondaryCheckBox:SetChecked(ChatWindowSettingsWindow.ChatWindow._settings.ChatGroups[ac].KinNoAnnounce);
              end
              secondaryCheckBox:SetVisible(true);
              ---@diagnostic disable-next-line: duplicate-set-field
              secondaryCheckBox.CheckedChanged = function(cb)
                if cb.SelfLootNoCoins then
                  ChatWindowSettingsWindow.ChatWindow._settings.ChatGroups[ac].SelfLootNoCoins = cb:IsChecked();
                elseif cb.KinNoAnnounce then
                  ChatWindowSettingsWindow.ChatWindow._settings.ChatGroups[ac].KinNoAnnounce = cb:IsChecked();
                end
                Save();
              end

              --[[ --No point to this really, plugin loads after MOTD is displayed anyway
              if chatType == Turbine.ChatType.Kinship then
                local thirdCheckBox = Turbine.UI.Lotro.CheckBox();
                thirdCheckBox.KinNoMOTD = chatType == Turbine.ChatType.Kinship;
                thirdCheckBox:SetFont(Turbine.UI.Lotro.Font.TrajanPro18);
                thirdCheckBox:SetParent(wrapper);
                thirdCheckBox:SetLeft(secondaryCheckBox:GetLeft() + secondaryCheckBox:GetWidth());
                thirdCheckBox:SetTop(checkBox:GetTop());
                thirdCheckBox:SetHeight(22);
                thirdCheckBox:SetWidth(120);
                if thirdCheckBox.KinNoMOTD then
                  thirdCheckBox:SetText(" " .. "No MOTD");
                  thirdCheckBox:SetChecked(ChatWindowSettingsWindow.ChatWindow._settings.ChatGroups[ac].KinNoMOTD);
                end
                thirdCheckBox:SetVisible(true);
                ---@diagnostic disable-next-line: duplicate-set-field
                thirdCheckBox.CheckedChanged = function(cb)
                  if cb.KinNoMOTD then
                    ChatWindowSettingsWindow.ChatWindow._settings.ChatGroups[ac].KinNoMOTD = cb:IsChecked();
                  end
                  Save();
                end
              end
              ]]
            end

            if loopct > 1 then
              local spacerI = Turbine.UI.Label();
              spacerI:SetParent(wrapper);
              spacerI:SetSize(ChatWindowSettingsWindow.ChannelListBox:GetWidth(), 1);
              spacerI:SetBackColor(Turbine.UI.Color.Gray);
              spacerI:SetTop(0);
            end

            ChatWindowSettingsWindow.ChannelListBox:AddItem(wrapper);

            loopct = loopct + 1;

          end
        end

        ChatWindowSettingsWindow.NameLabel = Turbine.UI.Label();
        ChatWindowSettingsWindow.NameLabel:SetFont(Turbine.UI.Lotro.Font.TrajanPro18);
        ChatWindowSettingsWindow.NameLabel:SetParent(ChatWindowSettingsWindow);
        ChatWindowSettingsWindow.NameLabel:SetText("Name");
        ChatWindowSettingsWindow.NameLabel:SetLeft(15);
        ChatWindowSettingsWindow.NameLabel:SetTop(50);
        ChatWindowSettingsWindow.NameLabel:SetWidth(140);
        ChatWindowSettingsWindow.NameLabel:SetVisible(true);

        ChatWindowSettingsWindow.NameTextBox = Turbine.UI.Lotro.TextBox();
        ChatWindowSettingsWindow.NameTextBox:SetFont(Turbine.UI.Lotro.Font.TrajanPro18);
        ChatWindowSettingsWindow.NameTextBox:SetParent(ChatWindowSettingsWindow);
        ChatWindowSettingsWindow.NameTextBox:SetLeft((ChatWindowSettingsWindow.NameLabel:GetLeft() + ChatWindowSettingsWindow.NameLabel:GetWidth()));
        ChatWindowSettingsWindow.NameTextBox:SetTop(ChatWindowSettingsWindow.NameLabel:GetTop() - 2);
        ChatWindowSettingsWindow.NameTextBox:SetWidth((ChatWindowSettingsWindow:GetWidth() / 2) - ChatWindowSettingsWindow.NameLabel:GetWidth());
        ChatWindowSettingsWindow.NameTextBox:SetHeight(20);
        ChatWindowSettingsWindow.NameTextBox:SetVisible(true);
        ChatWindowSettingsWindow.NameTextBox:SetText(ChatWindowSettingsWindow.ChatWindow._settings.Name);
        ChatWindowSettingsWindow.NameTextBox:SetMultiline(false);
        ChatWindowSettingsWindow.NameTextBox.OriginalName = ChatWindowSettingsWindow.ChatWindow._settings.Name;
        ---@diagnostic disable-next-line: duplicate-set-field
        ChatWindowSettingsWindow.NameTextBox.FocusLost = function(ssn,ssa)
          local exists = false;
          for key,value in pairs(_G.WindowList) do
            if _G.WindowList[key]._settings.Name == ssn:GetText()
               and ssn:GetText() ~= ChatWindowSettingsWindow.NameTextBox.OriginalName
            then
              exists = true;
              Turbine.Shell.WriteLine("Window '" .. ssn:GetText() .. "' already exists")
              ChatWindowSettingsWindow.NameTextBox:SetText(ChatWindowSettingsWindow.ChatWindow._settings.Name);
            end
          end
          if not exists then
            ChatWindowSettingsWindow.ChatWindow._settings.Name = ssn:GetText();
            ChatWindowSettingsWindow.ChatWindow.Name:SetText(ssn:GetText());
            ChatWindowSettingsWindow.NameTextBox.OriginalName = ssn:GetText();
          end
        end

        ChatWindowSettingsWindow.HoverOpacityLabel = Turbine.UI.Label();
        ChatWindowSettingsWindow.HoverOpacityLabel:SetFont(Turbine.UI.Lotro.Font.TrajanPro18);
        ChatWindowSettingsWindow.HoverOpacityLabel:SetParent(ChatWindowSettingsWindow);
        ChatWindowSettingsWindow.HoverOpacityLabel:SetText("Hover Opacity");
        ChatWindowSettingsWindow.HoverOpacityLabel:SetLeft(15);
        ChatWindowSettingsWindow.HoverOpacityLabel:SetTop(ChatWindowSettingsWindow.NameLabel:GetTop() + 30);
        ChatWindowSettingsWindow.HoverOpacityLabel:SetWidth(ChatWindowSettingsWindow.NameLabel:GetWidth());
        ChatWindowSettingsWindow.HoverOpacityLabel:SetVisible(true);

        ChatWindowSettingsWindow.HoverOpacityZeroLabel = Turbine.UI.Label();
        ChatWindowSettingsWindow.HoverOpacityZeroLabel:SetFont(Turbine.UI.Lotro.Font.TrajanPro14);
        ChatWindowSettingsWindow.HoverOpacityZeroLabel:SetForeColor(Turbine.UI.Color.Gold);
        ChatWindowSettingsWindow.HoverOpacityZeroLabel:SetParent(ChatWindowSettingsWindow);
        ChatWindowSettingsWindow.HoverOpacityZeroLabel:SetText("0");
        ChatWindowSettingsWindow.HoverOpacityZeroLabel:SetLeft((ChatWindowSettingsWindow.HoverOpacityLabel:GetLeft() + ChatWindowSettingsWindow.HoverOpacityLabel:GetWidth()));
        ChatWindowSettingsWindow.HoverOpacityZeroLabel:SetTop(ChatWindowSettingsWindow.HoverOpacityLabel:GetTop() + 3);
        ChatWindowSettingsWindow.HoverOpacityZeroLabel:SetWidth(7);
        ChatWindowSettingsWindow.HoverOpacityZeroLabel:SetVisible(true);

        local chatHoverOpacityValue = math.floor(ChatWindowSettingsWindow.ChatWindow._settings.ChatHoverOpacity * 255);
        ChatWindowSettingsWindow.HoverOpacitySlider = Turbine.UI.Lotro.ScrollBar();
        ChatWindowSettingsWindow.HoverOpacitySlider:SetOrientation(Turbine.UI.Orientation.Horizontal);
        ChatWindowSettingsWindow.HoverOpacitySlider:SetBackColor(Turbine.UI.Color(0.9,0,0,0));
        ChatWindowSettingsWindow.HoverOpacitySlider:SetParent(ChatWindowSettingsWindow);
        ChatWindowSettingsWindow.HoverOpacitySlider:SetLeft((ChatWindowSettingsWindow.HoverOpacityZeroLabel:GetLeft() + ChatWindowSettingsWindow.HoverOpacityZeroLabel:GetWidth()) + 10);
        ChatWindowSettingsWindow.HoverOpacitySlider:SetTop(ChatWindowSettingsWindow.HoverOpacityLabel:GetTop() + 4);
        ChatWindowSettingsWindow.HoverOpacitySlider:SetWidth(213);
        ChatWindowSettingsWindow.HoverOpacitySlider:SetHeight(10);
        ChatWindowSettingsWindow.HoverOpacitySlider:SetMinimum(0);
        ChatWindowSettingsWindow.HoverOpacitySlider:SetMaximum(255);
        ChatWindowSettingsWindow.HoverOpacitySlider:SetSmallChange(1);
        ChatWindowSettingsWindow.HoverOpacitySlider:SetLargeChange(10);
        ChatWindowSettingsWindow.HoverOpacitySlider:SetValue(chatHoverOpacityValue);
        ChatWindowSettingsWindow.HoverOpacitySlider:SetVisible(true);
        ---@diagnostic disable-next-line: duplicate-set-field
        ChatWindowSettingsWindow.HoverOpacitySlider.ValueChanged = function(sss,ssa)
          if false then
            DebugWrite("slider: " ..tostring(ChatWindowSettingsWindow.HoverOpacitySlider:GetValue()));
            DebugWrite("slider / 255: " .. tostring(ChatWindowSettingsWindow.HoverOpacitySlider:GetValue() * 0.003921568));
            DebugWrite("setting: " .. tostring(ChatWindowSettingsWindow.ChatWindow._settings.ChatHoverOpacity));
            DebugWrite("setting * 255: " .. tostring(math.floor(ChatWindowSettingsWindow.ChatWindow._settings.ChatHoverOpacity * 255)));
          end
          local chatHoverOpacityValue = ChatWindowSettingsWindow.HoverOpacitySlider:GetValue() * 0.003921568;
          if chatHoverOpacityValue < ChatWindowSettingsWindow.ChatWindow._settings.ChatOpacity then
            ChatWindowSettingsWindow.HoverOpacitySlider:SetValue(ChatWindowSettingsWindow.ChatWindow._settings.ChatOpacity * 255);
          else
            ChatWindowSettingsWindow.ChatWindow._settings.ChatHoverOpacity = chatHoverOpacityValue;
          end
        end

        ChatWindowSettingsWindow.HoverOpacitySolidLabel = Turbine.UI.Label();
        ChatWindowSettingsWindow.HoverOpacitySolidLabel:SetFont(Turbine.UI.Lotro.Font.TrajanPro14);
        ChatWindowSettingsWindow.HoverOpacitySolidLabel:SetForeColor(Turbine.UI.Color.Gold);
        ChatWindowSettingsWindow.HoverOpacitySolidLabel:SetParent(ChatWindowSettingsWindow);
        ChatWindowSettingsWindow.HoverOpacitySolidLabel:SetText("255");
        ChatWindowSettingsWindow.HoverOpacitySolidLabel:SetLeft((ChatWindowSettingsWindow.HoverOpacitySlider:GetLeft() + ChatWindowSettingsWindow.HoverOpacitySlider:GetWidth()) + 10);
        ChatWindowSettingsWindow.HoverOpacitySolidLabel:SetTop(ChatWindowSettingsWindow.HoverOpacityLabel:GetTop() + 3);
        ChatWindowSettingsWindow.HoverOpacitySolidLabel:SetWidth(40);
        ChatWindowSettingsWindow.HoverOpacitySolidLabel:SetVisible(true);


        ChatWindowSettingsWindow.PurgeAfterLabel = Turbine.UI.Label();
        ChatWindowSettingsWindow.PurgeAfterLabel:SetFont(Turbine.UI.Lotro.Font.TrajanPro18);
        ChatWindowSettingsWindow.PurgeAfterLabel:SetParent(ChatWindowSettingsWindow);
        ChatWindowSettingsWindow.PurgeAfterLabel:SetText("Purge After");
        ChatWindowSettingsWindow.PurgeAfterLabel:SetLeft(15);
        ChatWindowSettingsWindow.PurgeAfterLabel:SetTop(ChatWindowSettingsWindow.HoverOpacityLabel:GetTop() + 30);
        ChatWindowSettingsWindow.PurgeAfterLabel:SetWidth(ChatWindowSettingsWindow.NameLabel:GetWidth());
        ChatWindowSettingsWindow.PurgeAfterLabel:SetVisible(true);

        local purgeAfterMinutes = tostring(ChatWindowSettingsWindow.ChatWindow._settings.PurgeChatMinutes);
        if ChatWindowSettingsWindow.ChatWindow._settings.PurgeChatMinutes == nil then
          ChatWindowSettingsWindow.ChatWindow._settings.PurgeChatMinutes = 0;
        end
        if ChatWindowSettingsWindow.ChatWindow._settings.PurgeChatMinutes < 1 then
          purgeAfterMinutes = "";
        end
        ChatWindowSettingsWindow.PurgeAfterTextBox = Turbine.UI.Lotro.TextBox();
        ChatWindowSettingsWindow.PurgeAfterTextBox:SetFont(Turbine.UI.Lotro.Font.TrajanPro18);
        ChatWindowSettingsWindow.PurgeAfterTextBox:SetParent(ChatWindowSettingsWindow);
        ChatWindowSettingsWindow.PurgeAfterTextBox:SetLeft((ChatWindowSettingsWindow.PurgeAfterLabel:GetLeft() + ChatWindowSettingsWindow.PurgeAfterLabel:GetWidth() + 1));
        ChatWindowSettingsWindow.PurgeAfterTextBox:SetTop(ChatWindowSettingsWindow.PurgeAfterLabel:GetTop() - 2);
        ChatWindowSettingsWindow.PurgeAfterTextBox:SetWidth(50);
        ChatWindowSettingsWindow.PurgeAfterTextBox:SetHeight(20);
        ChatWindowSettingsWindow.PurgeAfterTextBox:SetVisible(true);
        ChatWindowSettingsWindow.PurgeAfterTextBox:SetText(purgeAfterMinutes);
        ChatWindowSettingsWindow.PurgeAfterTextBox:SetMultiline(false);
        ---@diagnostic disable-next-line: duplicate-set-field
        ChatWindowSettingsWindow.PurgeAfterTextBox.TextChanged = function(sss,ssa)
          local textValue = ChatWindowSettingsWindow.PurgeAfterTextBox:GetText();
          if textValue == nil or textValue == "" or tonumber("0"..textValue) < 1 then
            ChatWindowSettingsWindow.PurgeAfterTextBox:SetText("");
          else
            ChatWindowSettingsWindow.PurgeAfterTextBox:SetText(tostring(tonumber("0"..textValue)));
          end
        end
        ---@diagnostic disable-next-line: duplicate-set-field
        ChatWindowSettingsWindow.PurgeAfterTextBox.FocusLost = function(sss,ssa)
          local textValue = ChatWindowSettingsWindow.PurgeAfterTextBox:GetText();
          if textValue == nil or textValue == "" or tonumber("0"..textValue) < 1 then
            ChatWindowSettingsWindow.ChatWindow._settings.PurgeChatMinutes = 0;
          else
            ChatWindowSettingsWindow.ChatWindow._settings.PurgeChatMinutes = tonumber("0"..textValue);
          end
          if ChatWindowSettingsWindow.ChatWindow._settings.PurgeChatMinutes == nil then
            ChatWindowSettingsWindow.ChatWindow._settings.PurgeChatMinutes = 0;
          end
        end

        ChatWindowSettingsWindow.purgeAfterMinutesLabel = Turbine.UI.Label();
        ChatWindowSettingsWindow.purgeAfterMinutesLabel:SetFont(Turbine.UI.Lotro.Font.TrajanPro18);
        ChatWindowSettingsWindow.purgeAfterMinutesLabel:SetParent(ChatWindowSettingsWindow);
        ChatWindowSettingsWindow.purgeAfterMinutesLabel:SetText("minute(s)");
        ChatWindowSettingsWindow.purgeAfterMinutesLabel:SetLeft((ChatWindowSettingsWindow.PurgeAfterTextBox:GetLeft() + ChatWindowSettingsWindow.PurgeAfterTextBox:GetWidth()) + 10);
        ChatWindowSettingsWindow.purgeAfterMinutesLabel:SetTop(ChatWindowSettingsWindow.PurgeAfterLabel:GetTop());
        ChatWindowSettingsWindow.purgeAfterMinutesLabel:SetWidth(200);
        ChatWindowSettingsWindow.purgeAfterMinutesLabel:SetVisible(true);

        ChatWindowSettingsWindow.HoverActivate = Turbine.UI.Lotro.CheckBox();
        ChatWindowSettingsWindow.HoverActivate:SetFont(Turbine.UI.Lotro.Font.TrajanPro18);
        ChatWindowSettingsWindow.HoverActivate:SetParent(ChatWindowSettingsWindow);
        ChatWindowSettingsWindow.HoverActivate:SetLeft((ChatWindowSettingsWindow.NameLabel:GetLeft() + ChatWindowSettingsWindow.NameLabel:GetWidth() + 1));
        ChatWindowSettingsWindow.HoverActivate:SetHeight(20);
        ChatWindowSettingsWindow.HoverActivate:SetTop(ChatWindowSettingsWindow.PurgeAfterLabel:GetTop() + 30);
        ChatWindowSettingsWindow.HoverActivate:SetWidth((ChatWindowSettingsWindow:GetWidth() / 2));
        ChatWindowSettingsWindow.HoverActivate:SetText(" Activate Window on Hover");
        ChatWindowSettingsWindow.HoverActivate:SetChecked(ChatWindowSettingsWindow.ChatWindow._settings.HoverActivate);
        ChatWindowSettingsWindow.HoverActivate:SetVisible(true);

        ---@diagnostic disable-next-line: duplicate-set-field
        ChatWindowSettingsWindow.HoverActivate.CheckedChanged = function(cb)
          ChatWindowSettingsWindow.ChatWindow._settings.HoverActivate = cb:IsChecked();
          Save();
        end

        ChatWindowSettingsWindow.IncludeTimestamps = Turbine.UI.Lotro.CheckBox();
        ChatWindowSettingsWindow.IncludeTimestamps:SetFont(Turbine.UI.Lotro.Font.TrajanPro18);
        ChatWindowSettingsWindow.IncludeTimestamps:SetParent(ChatWindowSettingsWindow);
        ChatWindowSettingsWindow.IncludeTimestamps:SetLeft((ChatWindowSettingsWindow.NameLabel:GetLeft() + ChatWindowSettingsWindow.NameLabel:GetWidth() + 1) );
        ChatWindowSettingsWindow.IncludeTimestamps:SetHeight(20);
        ChatWindowSettingsWindow.IncludeTimestamps:SetTop(ChatWindowSettingsWindow.HoverActivate:GetTop() + 30);
        ChatWindowSettingsWindow.IncludeTimestamps:SetWidth(ChatWindowSettingsWindow:GetWidth() - 100);
        ChatWindowSettingsWindow.IncludeTimestamps:SetText(" Include Timestamps");
        ChatWindowSettingsWindow.IncludeTimestamps:SetChecked(ChatWindowSettingsWindow.ChatWindow._settings.IncludeTimestamps);
        ChatWindowSettingsWindow.IncludeTimestamps:SetVisible(true);
        ---@diagnostic disable-next-line: duplicate-set-field
        ChatWindowSettingsWindow.IncludeTimestamps.CheckedChanged = function(cb)
          ChatWindowSettingsWindow.ChatWindow._settings.IncludeTimestamps = cb:IsChecked();
          Save();
        end

        local colorSettings = { };
        table.insert(colorSettings, { Order = 1, Name = "WindowBackColor", Color = self._settings.WindowBackColor });
        table.insert(colorSettings, { Order = 2, Name = "ToolbarBackColor", Color = self._settings.ToolbarBackColor });
        table.insert(colorSettings, { Order = 3, Name = "ToolbarForeColor", Color = self._settings.ToolbarForeColor });
        table.insert(colorSettings, { Order = 4, Name = "ToggleColor", Color = self._settings.ToggleColor });
        table.insert(colorSettings, { Order = 5, Name = "WarningColor", Color = self._settings.WarningColor });
        table.insert(colorSettings, { Order = 6, Name = "DisabledColor", Color = self._settings.DisabledColor });

        local colorWrapper = Turbine.UI.Control();
        colorWrapper:SetParent(ChatWindowSettingsWindow);
        colorWrapper:SetLeft(ChatWindowSettingsWindow.NameTextBox:GetLeft() + ChatWindowSettingsWindow.NameTextBox:GetWidth() + 50);
        colorWrapper:SetTop(ChatWindowSettingsWindow.NameLabel:GetTop());
        colorWrapper:SetWidth((ChatWindowSettingsWindow:GetWidth() / 2) - 90);
        colorWrapper:SetHeight(ChatWindowSettingsWindow.IncludeTimestamps:GetTop() + ChatWindowSettingsWindow.IncludeTimestamps:GetHeight());

        local colorTop = 0;
        for key,value in spairs(colorSettings, function(t,aa,bb) return t[bb].Order > t[aa].Order end) do

          local wrapperBG = Turbine.UI.Control();
          wrapperBG.SettingsName = value.Name;
          wrapperBG:SetParent(colorWrapper);
          wrapperBG:SetLeft(32);
          wrapperBG:SetSize(colorWrapper:GetWidth() - wrapperBG:GetLeft(), 26);
          wrapperBG:SetTop(colorTop);

          local wrapper = Turbine.UI.Control();
          wrapper.SettingsName = value.Name;
          wrapper:SetParent(colorWrapper);
          wrapper:SetSize(wrapperBG:GetLeft(), 26);
          wrapper:SetLeft(0);
          wrapper:SetTop(colorTop);
          wrapper:SetBackColor(Turbine.UI.Color(0.8,0,0,0));
          colorTop = colorTop + wrapper:GetHeight();

          local colorPickerLabel = Turbine.UI.Label();
          colorPickerLabel.SettingsName = wrapper.SettingsName;
          colorPickerLabel:SetParent(wrapper);
          colorPickerLabel:SetBackColor( Turbine.UI.Color(value.Color.R, value.Color.G, value.Color.B));
          colorPickerLabel:SetForeColor(colorPickerLabel:GetBackColor());
          colorPickerLabel:SetTop(6);
          colorPickerLabel:SetLeft(6);
          colorPickerLabel:SetSize(20,15);
          colorPickerLabel:SetVisible(true);
          ---@diagnostic disable-next-line: duplicate-set-field
          colorPickerLabel.MouseClick = function(f,ar)
            ChatWindowSettingsWindow.ChatWindow.PickColor(f);
          end
          colorPickerLabel.ColorUpdated = function(lbl)
            ChatWindowSettingsWindow.ChatWindow._settings[lbl.SettingsName] = lbl:GetBackColor();
            lbl:SetForeColor(self._settings[colorPickerLabel.SettingsName]);
            ChatWindowSettingsWindow.ChatWindow:UpdateColors();
            Save();
          end

          local settingsLabel = Turbine.UI.Label();
          settingsLabel:SetFont(Turbine.UI.Lotro.Font.TrajanPro18);
          settingsLabel:SetParent(wrapperBG);
          settingsLabel:SetTop(5);
          settingsLabel:SetLeft(0);
          settingsLabel:SetHeight(22);
          settingsLabel:SetWidth(wrapperBG:GetWidth());
          settingsLabel:SetText("  " .. wrapper.SettingsName:gsub("(%l)(%u)", "%1 %2"));
          settingsLabel:SetVisible(true);

        end

        ChatWindowSettingsWindow.CloseButton = Turbine.UI.Lotro.GoldButton();
        ChatWindowSettingsWindow.CloseButton:SetParent(ChatWindowSettingsWindow);
        ChatWindowSettingsWindow.CloseButton:SetText("Close Window");
        ChatWindowSettingsWindow.CloseButton:SetHeight(20);
        ChatWindowSettingsWindow.CloseButton:SetWidth(150);
        ChatWindowSettingsWindow.CloseButton:SetTop(ChatWindowSettingsWindow:GetHeight() - 40);
        ChatWindowSettingsWindow.CloseButton:SetLeft((ChatWindowSettingsWindow:GetWidth() / 2) - ChatWindowSettingsWindow.CloseButton:GetWidth() / 2);
        ChatWindowSettingsWindow.CloseButton:SetVisible(true);
        ---@diagnostic disable-next-line: duplicate-set-field
        ChatWindowSettingsWindow.CloseButton.Click = function(btn,arg)
          ChatWindowSettingsWindow:Close();
        end

        local spacerF = Turbine.UI.Label();
        spacerF:SetParent(ChatWindowSettingsWindow);
        spacerF:SetLeft(10);
        spacerF:SetSize(ChatWindowSettingsWindow:GetWidth() - 20, 1);
        spacerF:SetBackColor(Turbine.UI.Color.Gray);
        spacerF:SetTop(ChatWindowSettingsWindow.CloseButton:GetTop() - 24);


        ChatWindowSettingsWindow.ChannelGroupsListBox = Turbine.UI.ListBox();
        ChatWindowSettingsWindow.ChannelGroupsListBox:SetParent(ChatWindowSettingsWindow);
        ChatWindowSettingsWindow.ChannelGroupsListBox:SetLeft(ChatWindowSettingsWindow.NameLabel:GetLeft());
        ChatWindowSettingsWindow.ChannelGroupsListBox:SetTop(colorTop + 90);
        ChatWindowSettingsWindow.ChannelGroupsListBox:SetHeight((ChatWindowSettingsWindow.CloseButton:GetTop() - 10) - ((ChatWindowSettingsWindow.ChannelGroupsListBox:GetTop() + 50)));
        ChatWindowSettingsWindow.ChannelGroupsListBox:SetWidth(300);
        ---@diagnostic disable-next-line: duplicate-set-field
        ChatWindowSettingsWindow.ChannelGroupsListBox.SelectedIndexChanged = function(sc,ac)
          for i=ChatWindowSettingsWindow.ChannelGroupsListBox:GetItemCount(),1,-1 do
            ChatWindowSettingsWindow.ChannelGroupsListBox:GetItem(i):SetBackColor(Turbine.UI.Color(0.9,0,0,0));
          end
          --sc:GetSelectedItem():SetBackColor(Turbine.UI.Color(0.9,0.53,0.53,0.35));
          sc:GetSelectedItem():SetBackColor(Turbine.UI.Color(0.9,0.15,0.15,0.15));
          ChatWindowSettingsWindow:LoadChatGroup((sc:GetSelectedItem()).ChatGroupId);
        end

        ChatWindowSettingsWindow.ChannelGroupsScrollbar = Turbine.UI.Lotro.ScrollBar();
        ChatWindowSettingsWindow.ChannelGroupsScrollbar:SetOrientation(Turbine.UI.Orientation.Vertical);
        ChatWindowSettingsWindow.ChannelGroupsScrollbar:SetParent(ChatWindowSettingsWindow);
        ChatWindowSettingsWindow.ChannelGroupsScrollbar:SetPosition(ChatWindowSettingsWindow.ChannelGroupsListBox:GetLeft() + ChatWindowSettingsWindow.ChannelGroupsListBox:GetWidth() + 1 ,ChatWindowSettingsWindow.ChannelGroupsListBox:GetTop());
        ChatWindowSettingsWindow.ChannelGroupsScrollbar:SetHeight(ChatWindowSettingsWindow.ChannelGroupsListBox:GetHeight());
        ChatWindowSettingsWindow.ChannelGroupsScrollbar:SetWidth(10);
        ChatWindowSettingsWindow.ChannelGroupsScrollbar:SetZOrder(10);

        ChatWindowSettingsWindow.ChannelGroupsListBox:SetVerticalScrollBar(ChatWindowSettingsWindow.ChannelGroupsScrollbar);

        ChatWindowSettingsWindow.ChannelListBox = Turbine.UI.ListBox();
        ChatWindowSettingsWindow.ChannelListBox:SetParent(ChatWindowSettingsWindow);
        ChatWindowSettingsWindow.ChannelListBox:SetLeft(340);
        ChatWindowSettingsWindow.ChannelListBox:SetTop(colorTop + 90);
        ChatWindowSettingsWindow.ChannelListBox:SetHeight((ChatWindowSettingsWindow.CloseButton:GetTop() - 10) - ((ChatWindowSettingsWindow.ChannelListBox:GetTop() + 15)));
        ChatWindowSettingsWindow.ChannelListBox:SetWidth((ChatWindowSettingsWindow:GetWidth() - ChatWindowSettingsWindow.ChannelListBox:GetLeft()) - 30);


        ChatWindowSettingsWindow.ChannelScrollbar = Turbine.UI.Lotro.ScrollBar();
        ChatWindowSettingsWindow.ChannelScrollbar:SetOrientation(Turbine.UI.Orientation.Vertical);
        ChatWindowSettingsWindow.ChannelScrollbar:SetParent(ChatWindowSettingsWindow);
        ChatWindowSettingsWindow.ChannelScrollbar:SetPosition(ChatWindowSettingsWindow.ChannelListBox:GetLeft() + ChatWindowSettingsWindow.ChannelListBox:GetWidth() + 1 ,ChatWindowSettingsWindow.ChannelListBox:GetTop());
        ChatWindowSettingsWindow.ChannelScrollbar:SetHeight(ChatWindowSettingsWindow.ChannelListBox:GetHeight());
        ChatWindowSettingsWindow.ChannelScrollbar:SetWidth(10);
        ChatWindowSettingsWindow.ChannelScrollbar:SetZOrder(10);

        ChatWindowSettingsWindow.ChannelListBox:SetVerticalScrollBar(ChatWindowSettingsWindow.ChannelScrollbar);

        ChatWindowSettingsWindow.ChannelScrollbar:SetVisible(true);
        ChatWindowSettingsWindow.ChannelListBox:SetVisible(true);


        local channelsHelpHeader = Turbine.UI.Label();
        channelsHelpHeader:SetParent(ChatWindowSettingsWindow);
        channelsHelpHeader:SetFont(Turbine.UI.Lotro.Font.TrajanPro14);
        channelsHelpHeader:SetMultiline(true);
        channelsHelpHeader:SetSize(ChatWindowSettingsWindow.ChannelListBox:GetWidth() - 10, 30);
        channelsHelpHeader:SetTop(ChatWindowSettingsWindow.ChannelListBox:GetTop() - 30)
        channelsHelpHeader:SetLeft(ChatWindowSettingsWindow.ChannelListBox:GetLeft());
        channelsHelpHeader:SetTextAlignment(Turbine.UI.ContentAlignment.TopLeft);
        channelsHelpHeader:SetText(" Check the channels you want this chat window to show.\n Click the color to change the channel's color.");

        local spacer = Turbine.UI.Label();
        spacer:SetParent(ChatWindowSettingsWindow);
        spacer:SetSize(ChatWindowSettingsWindow.ChannelListBox:GetWidth(), 1);
        spacer:SetBackColor(Turbine.UI.Color.Gray);
        spacer:SetTop(channelsHelpHeader:GetTop() + channelsHelpHeader:GetHeight());
        spacer:SetLeft(ChatWindowSettingsWindow.ChannelListBox:GetLeft());

        ChatWindowSettingsWindow:LoadGroups();

        ChatWindowSettingsWindow:SetText(self._settings.Name);

        ---@diagnostic disable-next-line: duplicate-set-field
        ChatWindowSettingsWindow.Closing = function(sender,args)
          Save();
        end

        ChatWindowSettingsWindow:SetVisible(true);
      end


      ---------------------------------------------------------------------------
      -- DISABLE MENU ITEM ------------------------------------------------------
      ---------------------------------------------------------------------------
      local menuDisable = Turbine.UI.MenuItem("Disable", true);
      ---@diagnostic disable-next-line: duplicate-set-field
      menuDisable.Click = function(s,a)
        self._settings.Active = false;
        self:Close();
        Save();
      end


      contextMenu:GetItems():Add(menuSettings);
      contextMenu:GetItems():Add(menuDisable);

      contextMenu:ShowMenu();

    end

  end

  ---@diagnostic disable-next-line: duplicate-set-field
  self.Toolbar.MouseDoubleClick = function(sender,args)
    self:ToggleMinMax();
  end

------------------------------------------ CHAT GROUPS LIST --------------------------------------------------------

  self.ChatGroups = Turbine.UI.ListBox();
  self.ChatGroups:SetParent(self.Toolbar);
  self.ChatGroups:SetOrientation(Turbine.UI.Orientation.Horizontal);
  self.ChatGroups:SetZOrder(1);
  self.ChatGroups:SetHeight(self._groupHeight);
  self.ChatGroups:SetWidth(self.Toolbar:GetWidth());
---@diagnostic disable-next-line: duplicate-set-field
  self.ChatGroups.MouseClick = function(sender,args)
    self.Toolbar:MouseClick(args);
  end

  self.ChatGroups.SetActiveChatGroup = function(s)
    for i=1,self.ChatGroups:GetItemCount() do
      local item = self.ChatGroups:GetItem(i);
      if item.ChatGroupId > -1 then
        if item.ChatGroupId == self._settings.ActiveChatGroupId then
          self._settings.ChatGroups[self._settings.ActiveChatGroupId].HasUnreadMessages = false;
          item:SetForeColor( Turbine.UI.Color(1, self._settings.ToggleColor.R, self._settings.ToggleColor.G, self._settings.ToggleColor.B));
          item:SetText(item.Name);
          local W, H = Libraries.Text:getSize( item:GetText(), Turbine.UI.Lotro.Font.Verdana16, "Auto" );
          item:SetWidth(W + 10);
        else
          item:SetForeColor( Turbine.UI.Color(1, self._settings.ToolbarForeColor.R, self._settings.ToolbarForeColor.G, self._settings.ToolbarForeColor.B));
        end
      end
    end
  end

  self.ChatGroups.UpdateGroups = function(s)

    self.ChatGroups:ClearItems();

    local loopct = 0;
    for key,value in spairs(self._settings.ChatGroups, function(t,aa,bb) return t[bb].Order > t[aa].Order end) do

      loopct = loopct + 1;
      if self._settings.ActiveChatGroupId == -1 then
        self._settings.ActiveChatGroupId = self._settings.ChatGroups[key].Id;
        --DebugWrite("self._settings.ActiveChatGroupId (B): " .. tostring(self._settings.ActiveChatGroupId));
      end
      local groupLabel = Turbine.UI.Label();
      groupLabel.Name = value.Name;
      groupLabel.MissedText = "";
      groupLabel.ChatGroupId = value.Id;
      if groupLabel.ChatGroupId == self._settings.ActiveChatGroupId then
        --DebugWrite("self._settings.ActiveChatGroupId (Y): "..value.Name.." - " .. tostring(value.Id) .. " - Active: " .. tostring(self._settings.ActiveChatGroupId));
        groupLabel:SetForeColor( Turbine.UI.Color(1, self._settings.ToggleColor.R, self._settings.ToggleColor.G, self._settings.ToggleColor.B));      
      else
        --DebugWrite("self._settings.ActiveChatGroupId (G): "..value.Name.." - " .. tostring(value.Id) .. " - Active: " .. tostring(self._settings.ActiveChatGroupId));
        groupLabel:SetForeColor( Turbine.UI.Color(1, self._settings.ToolbarForeColor.R, self._settings.ToolbarForeColor.G, self._settings.ToolbarForeColor.B));
      end
      groupLabel:SetFont(Turbine.UI.Lotro.Font.Verdana16);

      local c, i = self.ChatWrapper:GetChatWindowsById(value.Id);
      local append = iif(value.HasUnreadMessages and c:GetTextLength() > 0, "*", "");
      groupLabel:SetText(self._settings.ChatGroups[key].Name .. append);

      local W, H = Libraries.Text:getSize( groupLabel:GetText(), Turbine.UI.Lotro.Font.Verdana16, "Auto" );
      groupLabel:SetWidth(W + 10);
      groupLabel:SetHeight(self._groupHeight);
      groupLabel:SetFontStyle(Turbine.UI.FontStyle.Outline);
      groupLabel:SetOutlineColor(Turbine.UI.Color.Black);


      ---@diagnostic disable-next-line: duplicate-set-field
      groupLabel.MouseHover = function(ss,aa)
        self:OnHover();
      end

      ---@diagnostic disable-next-line: duplicate-set-field
      groupLabel.MouseEnter = function(ss,aa)
        if ss.ChatGroupId ~= self._settings.ActiveChatGroupId then
          ss:SetForeColor( Turbine.UI.Color(1, self._settings.ToggleColor.R, self._settings.ToggleColor.G, self._settings.ToggleColor.B));
        end
      end

      ---@diagnostic disable-next-line: duplicate-set-field
      groupLabel.MouseLeave = function(ss,aa)
        if ss.ChatGroupId ~= self._settings.ActiveChatGroupId then
          if ss.ChatGroupId == self._settings.ActiveChatGroupId then
            ss:SetForeColor( Turbine.UI.Color(1, self._settings.ToggleColor.R, self._settings.ToggleColor.G, self._settings.ToggleColor.B));
          else
            ss:SetForeColor( Turbine.UI.Color(1, self._settings.ToolbarForeColor.R, self._settings.ToolbarForeColor.G, self._settings.ToolbarForeColor.B));
          end
        end
      end

      ---@diagnostic disable-next-line: duplicate-set-field
      groupLabel.MouseClick = function(s,a)
        self.ChatWrapper:SetActiveChatGroup(s.ChatGroupId);
      end
      self.ChatGroups:AddItem(groupLabel);
    end

    self.ChatGroups:UpdateLayout();
  end

  self.ChatGroups.UpdateLayout = function(s)

    self.ChatGroups:SetVisible(self.ChatGroups:GetItemCount() > 1 and self._settings.Maximized and self.PinToolbar:IsVisible());

    if self._settings.ToolbarBottom then
      self.ChatGroups:SetTop(0);
    else
      self.ChatGroups:SetTop(self._headerHeight);
    end

    for i=1,self.ChatGroups:GetItemCount() do
      local item = self.ChatGroups:GetItem(i);
      item:SetVisible(self.ChatGroups:IsVisible());
      if item.ChatGroupId == -1 then
        if item.UpdateButton ~= nil then
          item:UpdateButton();
        end
      else
        if self._settings.ToolbarBottom then
          item:SetTextAlignment(Turbine.UI.ContentAlignment.TopCenter);
        else
          item:SetTextAlignment(Turbine.UI.ContentAlignment.BottomCenter);
        end
      end
    end

  end

  self.ChatGroups.SetMissedChat = function(s,a)
    self._settings.ChatGroups[a].HasUnreadMessages = true;
    self.ChatGroups:UpdateGroups();
  end


  ------------------------------------------LEFT SIDE BUTTONS --------------------------------------------------------

  self.MinMax = Turbine.UI.Label();
  self.MinMax:SetParent(self.Toolbar);
  self.MinMax:SetPosition(self._leftButtonPos, -2)
  self.MinMax:SetFontStyle(Turbine.UI.FontStyle.Outline);
  self.MinMax:SetMouseVisible(true);
  self.MinMax:SetOpacity(self._settings.ChatOpacity)
  self.MinMax:SetWidth(14);
  self.MinMax:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
  self.MinMax:SetZOrder(1);
  self.MinMax.UpdateButton = function()

    if self._settings.Maximized then
      self.MinMax:SetFont(Turbine.UI.Lotro.Font.Verdana14);
      self.MinMax:SetTop(self:GetBaseToolbarButtonTop()-2)
      self.MinMax:SetHeight(self._headerHeight);
      self.MinMax.TipText = "Collapse window\nRight-click to ";
      self.MinMax:SetText('_');
    else
      self.MinMax:SetFont(Turbine.UI.Lotro.Font.Verdana23);
      if self._settings.ToolbarPinned then
        self.MinMax:SetTop(self:GetBaseToolbarButtonTop()-7);
      else
        if self.ChatGroups:GetItemCount() > 1 then
          self.MinMax:SetTop(10);
        else
          self.MinMax:SetTop(self:GetBaseToolbarButtonTop()-7);
        end
      end
      self.MinMax:SetHeight(self._headerHeight + 8);
      self.MinMax.TipText = "Expand " .. self._settings.Name .. " window\nRight-click to ";
      self.MinMax:SetText('□');
    end

    if self._settings.Enabled then
      self.MinMax.TipText = self.MinMax.TipText .. "disable receiving messages";
      self.MinMax:SetForeColor(Turbine.UI.Color(1, self._settings.ToolbarForeColor.R, self._settings.ToolbarForeColor.G, self._settings.ToolbarForeColor.B));
    else
      self.MinMax.TipText = self.MinMax.TipText .. "enable receiving messages";
      self.MinMax:SetForeColor(Turbine.UI.Color(1, self._settings.DisabledColor.R, self._settings.DisabledColor.G, self._settings.DisabledColor.B));
    end

    self.MinMax2:SetFont(self.MinMax:GetFont());
    self.MinMax2:SetTop(self.MinMax:GetTop())
    self.MinMax2:SetSize(self.MinMax:GetSize());
    self.MinMax2.TipText = self.MinMax.TipText;
    self.MinMax2:SetText(self.MinMax:GetText());
    self.MinMax2:SetForeColor(self.MinMax:GetForeColor());

  end

  ---@diagnostic disable-next-line: duplicate-set-field
  self.MinMax.MouseClick = function(sender,args)

    if ( args.Button == Turbine.UI.MouseButton.Right ) then
      self._settings.Enabled = not self._settings.Enabled;
    else
      if self._settings.Maximized then
        self._settings.CollapseRight = false;
      end
      self:ToggleMinMax();
    end

    self.MinMax:UpdateButton();

  end

  -- Increment left button position for next button(s) to position themselves
  self._leftButtonPos = self.MinMax:GetLeft() + self.MinMax:GetWidth();

  -------------------------------------------------------------------------------------------------------------------------------

  self.PinToolbar = Turbine.UI.Label();
  self.PinToolbar:SetParent(self.Toolbar);
  self.PinToolbar:SetPosition(self._leftButtonPos, self:GetBaseToolbarButtonTop()-1)
  self.PinToolbar:SetForeColor(Turbine.UI.Color(1, self._settings.ToolbarForeColor.R, self._settings.ToolbarForeColor.G, self._settings.ToolbarForeColor.B));
  self.PinToolbar:SetFont(Turbine.UI.Lotro.Font.Verdana14);
  self.PinToolbar:SetFontStyle(Turbine.UI.FontStyle.Outline);
  self.PinToolbar:SetSize(20, self._headerHeight + 4);
  self.PinToolbar:SetMouseVisible(true);
  self.PinToolbar:SetOpacity(self._settings.ChatOpacity);
  self.PinToolbar:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
  self.PinToolbar.UpdateButton = function()
    self.PinToolbar:SetTop(self:GetBaseToolbarButtonTop()-1);
    if self._settings.ToolbarPinned then
      self.PinToolbar.TipText = "Unpin toolbar";
      self.PinToolbar:SetText('‡');
    else
      self.PinToolbar.TipText = "Pin toolbar";
      self.PinToolbar:SetText('†');
    end
  end
  self.PinToolbar:UpdateButton();

  ---@diagnostic disable-next-line: duplicate-set-field
  self.PinToolbar.MouseClick = function(sender,args)

    self._settings.ToolbarPinned = not self._settings.ToolbarPinned;
    self.PinToolbar:UpdateButton();
    self.Toolbar:UpdateLayout();

  end

  -- Increment left button position for next button(s) to position themselves
  self._leftButtonPos = self.PinToolbar:GetLeft() + self.PinToolbar:GetWidth();

  -------------------------------------------------------------------------------------------------------------------------------

  self.BottomToolbar = Turbine.UI.Label();
  self.BottomToolbar:SetParent(self.Toolbar);
  self.BottomToolbar:SetPosition(self._leftButtonPos-2, self:GetBaseToolbarButtonTop()-2)
  self.BottomToolbar:SetForeColor(Turbine.UI.Color(1, self._settings.ToolbarForeColor.R, self._settings.ToolbarForeColor.G, self._settings.ToolbarForeColor.B));
  self.BottomToolbar:SetFont(Turbine.UI.Lotro.Font.Verdana14);
  self.BottomToolbar:SetFontStyle(Turbine.UI.FontStyle.Outline);
  self.BottomToolbar:SetSize(18, self._headerHeight + 4);
  self.BottomToolbar:SetMouseVisible(true);
  self.BottomToolbar:SetOpacity(self._settings.ChatOpacity)
  self.BottomToolbar.UpdateButton = function()
    if self._settings.ToolbarBottom then
      self.BottomToolbar.TipText = "Top toolbar";
      self.BottomToolbar:SetTop(self:GetBaseToolbarButtonTop()-9)
      self.BottomToolbar:SetHeight(32)
      self.BottomToolbar:SetText('…');
      self.BottomToolbar:SetTextAlignment(Turbine.UI.ContentAlignment.TopCenter);
    else
      self.BottomToolbar.TipText = "Bottom toolbar";
      self.BottomToolbar:SetTop(self:GetBaseToolbarButtonTop()-4)
      self.BottomToolbar:SetHeight(22)
      self.BottomToolbar:SetText('…');
      self.BottomToolbar:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
    end
  end
  self.BottomToolbar:UpdateButton();

  ---@diagnostic disable-next-line: duplicate-set-field
  self.BottomToolbar.MouseClick = function(sender,args)

    Libraries.Hint:Remove();
    self._settings.ToolbarBottom = not self._settings.ToolbarBottom;
    self.BottomToolbar:UpdateButton();
    self.Toolbar:UpdateLayout();

  end

  -- Increment left button position for next button(s) to position themselves
  self._leftButtonPos = self.BottomToolbar:GetLeft() + self.BottomToolbar:GetWidth();

  -------------------------------------------------------------------------------------------------------------------------------


  self.Brightness = Turbine.UI.Label();
  self.Brightness:SetParent(self.Toolbar);
  self.Brightness:SetPosition(self._leftButtonPos, self:GetBaseToolbarButtonTop()-3)
  self.Brightness:SetForeColor(Turbine.UI.Color(1, self._settings.ToolbarForeColor.R, self._settings.ToolbarForeColor.G, self._settings.ToolbarForeColor.B));
  self.Brightness:SetFont(Turbine.UI.Lotro.Font.Verdana18);
  self.Brightness:SetFontStyle(Turbine.UI.FontStyle.Outline);
  self.Brightness:SetSize(22, self._headerHeight);
  self.Brightness:SetMouseVisible(true);
  self.Brightness:SetOpacity(self._settings.ChatOpacity);
  self.Brightness:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
  self.Brightness.TipText = "Scroll to change background transparency\nCan also use left and right click\nHold alt to change toolbar\nHold shift to change window\nHold control for fine adjustment";
  self.Brightness:SetText('¤');

  self.Brightness.UpdateButton = function()
    self.Brightness:SetTop(self:GetBaseToolbarButtonTop()-3)
  end

    ---@diagnostic disable-next-line: duplicate-set-field
    self.Brightness.MouseClick = function(sender,args)
      local direction = 0;
      if args.Button == Turbine.UI.MouseButton.Right then
        direction = 1;
      end
      self.Brightness:MouseWheel({ Direction = direction });
    end

  ---@diagnostic disable-next-line: duplicate-set-field
  self.Brightness.MouseWheel = function(sender,args)

    local increment = 0.10;
    if self.HSize:IsControlKeyDown() then
      increment = 0.01;
    end

    if self.HSize:IsAltKeyDown() then
      if args.Direction < 1 then
        self._settings.ToolbarOpacity = self._settings.ToolbarOpacity - increment;
        if self._settings.ToolbarOpacity <= 0.0 then
          self._settings.ToolbarOpacity = 0.0;
        end
      else
        self._settings.ToolbarOpacity = settings.ToolbarOpacity + increment;
        if self._settings.ToolbarOpacity >= 1.0 then
          self._settings.ToolbarOpacity = 1.0;
        end
      end
      self.ToolbarBackground:SetBackColor(Turbine.UI.Color( self._settings.ToolbarOpacity, self._settings.ToolbarBackColor.R, self._settings.ToolbarBackColor.G, self._settings.ToolbarBackColor.B ));
      self.Name:UpdateVisibility();
    elseif self.HSize:IsShiftKeyDown() then
        if args.Direction < 1 then
          self._settings.WindowOpacity = self._settings.WindowOpacity - increment;
          if self._settings.WindowOpacity <= 0.0 then
            self._settings.WindowOpacity = 0.0;
          end
        else
          self._settings.WindowOpacity = settings.WindowOpacity + increment;
          if self._settings.WindowOpacity >= 1.0 then
            self._settings.WindowOpacity = 1.0;
          end
        end
        self:SetOpacity(self._settings.WindowOpacity);
    else
      local syncHover = math.floor(self._settings.ChatOpacity * 1000) == math.floor(self._settings.ChatHoverOpacity * 1000);
      if args.Direction < 1 then
        self._settings.ChatOpacity = self._settings.ChatOpacity - increment;
        if self._settings.ChatOpacity <= 0.0 then
          self._settings.ChatOpacity = 0.0;
        end
      else
        self._settings.ChatOpacity = settings.ChatOpacity + increment;
        if self._settings.ChatOpacity >= 1.0 then
          self._settings.ChatOpacity = 1.0;
        end
      end

      if syncHover then
        self._settings.ChatHoverOpacity = self._settings.ChatOpacity;
      end

      if self.HoverActive then
        self:OnHover();
      end
      self:SetBackColor(Turbine.UI.Color( self._settings.ChatOpacity, self._settings.WindowBackColor.R, self._settings.WindowBackColor.G, self._settings.WindowBackColor.B ));
    end

  end

  -- Increment left button position for next button(s) to position themselves
  self._leftButtonPos = self.Brightness:GetLeft() + self.Brightness:GetWidth();

  -------------------------------------------------------------------------------------------------------------------------------

  self.VSize = Turbine.UI.Label();
  self.VSize:SetParent(self.Toolbar);
  self.VSize:SetPosition(self._leftButtonPos - 3, self:GetBaseToolbarButtonTop()-1)
  self.VSize:SetForeColor(Turbine.UI.Color(1, self._settings.ToolbarForeColor.R, self._settings.ToolbarForeColor.G, self._settings.ToolbarForeColor.B));
  self.VSize:SetFont(Turbine.UI.Lotro.Font.Verdana14);
  self.VSize:SetFontStyle(Turbine.UI.FontStyle.Outline);
  self.VSize:SetSize(20, self._headerHeight);
  self.VSize:SetMouseVisible(true);
  self.VSize:SetOpacity(self._settings.ChatOpacity);
  self.VSize:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
  self.VSize.TipText = "Scroll to change height\nCan also use left and right click\nHold control for fine adjustment";
  self.VSize:SetText('+');

  self.VSize.UpdateButton = function()
    self.VSize:SetTop(self:GetBaseToolbarButtonTop()-1)
  end

  ---@diagnostic disable-next-line: duplicate-set-field
  self.VSize.MouseClick = function(sender,args)
    local direction = 0;
    if args.Button == Turbine.UI.MouseButton.Left then
      direction = 1;
    end
    self.VSize:MouseWheel({ Direction = direction });
  end

  ---@diagnostic disable-next-line: duplicate-set-field
  self.VSize.MouseWheel = function(sender,args)

    local increment = 16;
    if self.HSize:IsControlKeyDown() then
      increment = 1;
    end

    if args.Direction > 0 then
      if self._settings.ToolbarBottom then
        self._settings.Height = self._settings.Height + increment;
      else
        self._settings.Height = self._settings.Height - increment;
      end
    else
      if self._settings.ToolbarBottom then
        self._settings.Height = self._settings.Height - increment;
      else
        self._settings.Height = self._settings.Height + increment;
      end
    end

    if self._settings.Height < 64 then
      self._settings.Height  = 64;
    end

    if self._settings.ToolbarBottom then
      if args.Direction > 0 then
        self:SetTop(self:GetTop() - increment);
      else
        self:SetTop(self:GetTop() + increment);
      end
      self._settings.Top = self:GetTop();
    end

    self:SetHeight(self._settings.Height);
    self.ChatWrapper:SetChatWindowsHeight(self._settings.Height-(self:GetHeaderHeightWithChatGroups()));

    if self._settings.ToolbarPinned or self._settings.ToolbarBottom then
      self.Toolbar:UpdateLayout({ FromVSize = true });
    end

  end

  -- Increment left button position for next button(s) to position themselves
  self._leftButtonPos = self.VSize:GetLeft() + self.VSize:GetWidth();

  -------------------------------------------------------------------------------------------------------------------------------

  self.HSize = Turbine.UI.Label();
  self.HSize:SetParent(self.Toolbar);
  self.HSize:SetPosition(self._leftButtonPos, self:GetBaseToolbarButtonTop()-3)
  self.HSize:SetForeColor(Turbine.UI.Color(1, self._settings.ToolbarForeColor.R, self._settings.ToolbarForeColor.G, self._settings.ToolbarForeColor.B));
  self.HSize:SetFont(Turbine.UI.Lotro.Font.Verdana18);
  self.HSize:SetFontStyle(Turbine.UI.FontStyle.Outline);
  self.HSize:SetSize(20, self._headerHeight+2);
  self.HSize:SetMouseVisible(true);
  self.HSize:SetOpacity(self._settings.ChatOpacity);
  self.HSize:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
  self.HSize.TipText = "Scroll to change width\nCan also use left and right click\nHold control for fine adjustment";
  self.HSize:SetText('‹›');

  self.HSize.UpdateButton = function()
    self.HSize:SetTop(self:GetBaseToolbarButtonTop()-3)
  end

  ---@diagnostic disable-next-line: duplicate-set-field
  self.HSize.MouseClick = function(sender,args)
    local direction = 0;
    if args.Button == Turbine.UI.MouseButton.Left then
      direction = 1;
    end
    self.HSize:MouseWheel({ Direction = direction });
  end

  ---@diagnostic disable-next-line: duplicate-set-field
  self.HSize.MouseWheel = function(sender,args)

    local increment = 10;
    if self.HSize:IsControlKeyDown() then
      increment = 1;
    end

    local maxWidth = self._leftButtonPos + (self._settings.Width - self._rightButtonPos);

    if args.Direction > 0 then
      self._settings.Width = self._settings.Width + increment;
    else
      self._settings.Width = self._settings.Width - increment;
    end

    if self._settings.Width < maxWidth then
      self._settings.Width  = maxWidth;
    end

    self:SetWidth(self._settings.Width);

    self.Toolbar:SetSize(self:GetWidth(), self:GetHeaderHeightWithChatGroups());
    self.ToolbarBackground:SetSize(self.Toolbar:GetSize());

    self.ChatWrapper:SetChatWindowsWidth(self._settings.Width-self.ChatWrapper:GetChatWindowsLeft()-self._scrollbarWidth);

    self._rightButtonPos = self._settings.Width;

    self.MinMax2:SetLeft(self._rightButtonPos - self.MinMax2:GetWidth())
    self._rightButtonPos = self.MinMax2:GetLeft();

    self.Clear:SetLeft(self._rightButtonPos - self.Clear:GetWidth() - 2)
    self._rightButtonPos = self.Clear:GetLeft();

    self.Interactive:SetLeft(self._rightButtonPos - self.Interactive:GetWidth())
    self._rightButtonPos = self.Interactive:GetLeft();

    self.Name:SetWidth((self._settings.Width - self._leftButtonPos - (self._settings.Width - self._rightButtonPos)));
    self.Name:UpdateVisibility();

  end

  -- Increment left button position for next button(s) to position themselves
  self._leftButtonPos = self.HSize:GetLeft() + self.HSize:GetWidth();

  -------------------------------------------------------------------------------------------------------------------------------

  self.Tips = Turbine.UI.Label();
  self.Tips:SetParent(self.Toolbar);
  self.Tips:SetPosition(self._leftButtonPos, self:GetBaseToolbarButtonTop());
  self.Tips:SetForeColor(Turbine.UI.Color(1, self._settings.ToolbarForeColor.R, self._settings.ToolbarForeColor.G, self._settings.ToolbarForeColor.B));
  self.Tips:SetFont(Turbine.UI.Lotro.Font.Verdana14);
  self.Tips:SetFontStyle(Turbine.UI.FontStyle.Outline);
  self.Tips:SetSize(18, self._headerHeight);
  self.Tips:SetMouseVisible(true);
  self.Tips:SetOpacity(self._settings.ChatOpacity);
  self.Tips:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);

  self.Tips.UpdateText = function()
    if self._settings.TipsEnabled then
      self.Tips.TipText = "Click to disable tool tips";
      self.Tips:SetText('?');
    else
      self.Tips.TipText = "Click to enable tool tips"; --This won't actually show, but w/e
      self.Tips:SetText('¿');
    end
  end

  self.Tips.UpdateButton = function()
    self.Tips:SetTop(self:GetBaseToolbarButtonTop())
  end

  ---@diagnostic disable-next-line: duplicate-set-field
  self.Tips.MouseClick = function(sender,args)
    self._settings.TipsEnabled = not self._settings.TipsEnabled;
    self.Tips:UpdateText()
  end

  self.Tips:UpdateText();

  -- Increment left button position for next button(s) to position themselves
  self._leftButtonPos = self.Tips:GetLeft() + self.Tips:GetWidth();


  ---------------------------------------------------------------------------------------------------------------------
  ------------------------------------------END LEFT SIDE BUTTONS------------------------------------------------------
  ---------------------------------------------------------------------------------------------------------------------


  -- Center label for window Name, fills remainder between left and right buttons

  self.Name = Turbine.UI.Label();
  self.Name:SetParent(self.Toolbar);
  self.Name:SetForeColor(Turbine.UI.Color(1, self._settings.ToolbarForeColor.R, self._settings.ToolbarForeColor.G, self._settings.ToolbarForeColor.B));
  self.Name:SetFont(Turbine.UI.Lotro.Font.VerdanaBold16);
  self.Name:SetFontStyle(Turbine.UI.FontStyle.Outline);
  self.Name:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
  self.Name:SetMouseVisible(false);
  self.Name:SetOpacity(self._settings.ChatOpacity);
  self.Name:SetTop(self:GetBaseToolbarButtonTop()-2);
  self.Name:SetZOrder(-1);
  self.Name:SetLeft(self._leftButtonPos);
  self.Name:SetHeight(self._headerHeight+2);
  self.Name:SetText(self._settings.Name);
  self.Name:SetVisible(true);

  self.Name.UpdateButton = function()
    self.Name:SetTop(self:GetBaseToolbarButtonTop()-2)
  end

  self.Name.UpdateVisibility = function()

    --if self._settings.ToolbarOpacity < 0.5 then
      self.Name:SetFont(Turbine.UI.Lotro.Font.Verdana16);
    --else
    --  self.Name:SetFont(Turbine.UI.Lotro.Font.VerdanaBold16);
    --end
    self.Name:SetText(self._settings.Name);

    local W, H = Libraries.Text:getSize( self._settings.Name, Turbine.UI.Lotro.Font.VerdanaBold16, "Auto" );
    if self._settings.Width < (self._leftButtonPos + W + 4 + (self._settings.Width - self._rightButtonPos)) then
      self.Name:SetVisible(false);
    else
      self.Name:SetVisible(self.Tips:IsVisible());
    end

  end


  ---------------------------------------------------------------------------------------------------------------------
  ------------------------------------------RIGHT SIDE BUTTONS --------------------------------------------------------
  ---------------------------------------------------------------------------------------------------------------------


  self.MinMax2 = Turbine.UI.Label();
  self.MinMax2:SetParent(self.Toolbar);
  self.MinMax2:SetFontStyle(Turbine.UI.FontStyle.Outline);
  self.MinMax2:SetMouseVisible(true);
  self.MinMax2:SetOpacity(self._settings.ChatOpacity)
  self.MinMax2:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
  self.MinMax2:SetWidth(14);
  self.MinMax2:SetPosition(self._rightButtonPos - self.MinMax2:GetWidth(), self:GetBaseToolbarButtonTop()-2)
  self.MinMax2:SetZOrder(1);
  self.MinMax2.UpdateButton = self.MinMax.UpdateButton;


  ---@diagnostic disable-next-line: duplicate-set-field
  self.MinMax2.MouseClick = function(sender,args)

    if ( args.Button == Turbine.UI.MouseButton.Right ) then
      self._settings.Enabled = not self._settings.Enabled;
    else
      if self._settings.Maximized then
        self._settings.CollapseRight = true;
      end
      self:ToggleMinMax();
    end

    self.MinMax2.UpdateButton();

  end

  -- Increment right button position for next button(s) to position themselves
  self._rightButtonPos = self.MinMax2:GetLeft();


  -------------------------------------------------------------------------------------------------------------------------------


  self.Clear = Turbine.UI.Label();
  self.Clear:SetParent(self.Toolbar);
  self.Clear:SetForeColor(Turbine.UI.Color(1, self._settings.ToolbarForeColor.R, self._settings.ToolbarForeColor.G, self._settings.ToolbarForeColor.B));
  self.Clear:SetFont(Turbine.UI.Lotro.Font.Verdana12);
  self.Clear:SetFontStyle(Turbine.UI.FontStyle.Outline);
  self.Clear:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
  self.Clear:SetMouseVisible(true);
  self.Clear:SetOpacity(self._settings.ChatOpacity);
  self.Clear:SetSize(20, self._headerHeight + 4);
  self.Clear:SetPosition(self._rightButtonPos - self.Clear:GetWidth() - 4, self:GetBaseToolbarButtonTop()-2)
  self.Clear:SetText('Ø');
  self.Clear:SetZOrder(1);
  self.Clear.TipText = "Click to clear this window.\nRight-click to send 25 blank lines to Main LOTRO chat window.";

  self.Clear.UpdateButton = function()
    self.Clear:SetTop(self:GetBaseToolbarButtonTop()-2)
  end

  ---@diagnostic disable-next-line: duplicate-set-field
  self.Clear.MouseClick = function(sender,args)

    if ( args.Button == Turbine.UI.MouseButton.Right ) then
        -- Essentially clear the Main LOTRO chat window by sending a bunch of empty lines
        for i=1,25 do
          Turbine.Shell.WriteLine(" \n");
        end
    elseif ( args.Button == Turbine.UI.MouseButton.Left ) then
      -- Clear this window
      self:SetText('', self._settings.ActiveChatGroupId);
      self.Clear:SetForeColor(Turbine.UI.Color(1, self._settings.ToolbarForeColor.R, self._settings.ToolbarForeColor.G, self._settings.ToolbarForeColor.B));

      -- Empty out message logging tables
      for k,v in pairs(self._entityMessages) do
        self._entityMessages[k] = nil;
      end
      for k,v in pairs(self._entityMessageCounts) do
        self._entityMessageCounts[k] = nil;
      end
      for k,v in pairs(self._entityTypes) do
        self._entityTypes[k] = nil;
      end

    end

  end

  -- Increment right button position for next button(s) to position themselves
  self._rightButtonPos = self.Clear:GetLeft();

  -------------------------------------------------------------------------------------------------------------------------------

  self.Interactive = Turbine.UI.Label();
  self.Interactive:SetParent(self.Toolbar);
  self.Interactive:SetFont(Turbine.UI.Lotro.Font.Verdana16);
  self.Interactive:SetFontStyle(Turbine.UI.FontStyle.Outline);
  self.Interactive:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
  self.Interactive:SetMouseVisible(true);
  self.Interactive:SetOpacity(self._settings.ChatOpacity);
  self.Interactive:SetSize(22, self._headerHeight + 4);
  self.Interactive:SetPosition(self._rightButtonPos - self.Interactive:GetWidth(), self:GetBaseToolbarButtonTop()-2)
  self.Interactive:SetZOrder(1);

  self.Interactive.UpdateButton = function()
    self.Interactive:SetTop(self:GetBaseToolbarButtonTop()-2)
    if self._settings.ChatClickThrough then
      self.Interactive:SetForeColor( Turbine.UI.Color(1, self._settings.ToolbarForeColor.R, self._settings.ToolbarForeColor.G, self._settings.ToolbarForeColor.B));
    else
      self.Interactive:SetForeColor( Turbine.UI.Color(1, self._settings.ToggleColor.R, self._settings.ToggleColor.G, self._settings.ToggleColor.B) );
    end

    if self._settings.InteractiveMode then
      self.Interactive:SetText('≠');
    else
      self.Interactive:SetText('=');
    end

    local interactiveMode = "disable";
    if not self._settings.ChatClickThrough then
      interactiveMode = "enable";
    end
    self.Interactive.TipText = "Left click to show linkable chat.\nRight-click to " .. interactiveMode .. " mouse click-through.";    
  end

  self.Interactive:UpdateButton();

  ---@diagnostic disable-next-line: duplicate-set-field
  self.Interactive.MouseClick=function(sender,args)
    --DebugWrite("self.Interactive.MouseClick");
    if ( args.Button == Turbine.UI.MouseButton.Right ) then
      self._settings.ChatClickThrough = not self._settings.ChatClickThrough;
      self.ChatWrapper:SetChatWindowsMouseVisible(not self._settings.ChatClickThrough);

      local chatMode = "linkable";
      if self.ChatWrapper:GetInteractiveMode() then
        chatMode = "clean";
      end
      if self._settings.ChatClickThrough then
        self.Interactive:SetForeColor(Turbine.UI.Color(1, self._settings.ToolbarForeColor.R, self._settings.ToolbarForeColor.G, self._settings.ToolbarForeColor.B));
        self.Interactive.TipText = "Left click to show " .. chatMode .. " chat.\nRight-click to disable mouse click-through.";
      else
        self.Interactive:SetForeColor(Turbine.UI.Color(1, self._settings.ToggleColor.R, self._settings.ToggleColor.G, self._settings.ToggleColor.B));
        self.Interactive.TipText = "Left click to show " .. chatMode .. " chat.\nRight-click to enable mouse click-through.";
      end
    else
      local interactiveMode = "disable";
      if not self._settings.ChatClickThrough then
        interactiveMode = "enable";
      end
      self._settings.InteractiveMode = not self.ChatWrapper:GetInteractiveMode();
      if self.ChatWrapper:GetInteractiveMode() then
        self.Interactive.TipText = "Left click to show linkable chat.\nRight-click to " .. interactiveMode .. " mouse click-through.";
        self.Interactive:SetText('=');
        self.ChatWrapper:SetInteractiveMode(false);
      else
        self.Interactive.TipText = "Left click to show clean chat.\nRight-click to " .. interactiveMode .. " mouse click-through.";
        self.Interactive:SetText('≠');
        self.ChatWrapper:SetInteractiveMode(true);
      end
      self.Interactive:UpdateButton();
    end

  end

  -- Increment right button position for next button(s) to position themselves
  self._rightButtonPos = self.Interactive:GetLeft();


  ------------------------------------------END RIGHT SIDE BUTTONS------------------------------------------------------------------


  ------------------------------------------CLASS METHODS------------------------------------------------------------------

  function self:DoPurge()

    local purgeTime = Turbine.Engine.GetLocalTime() - (self._settings.PurgeChatMinutes * 60);
    local doPurge = false;

    for k,v in spairs(self._entityMessages, function(t,a,b) return t[b].LocalTime > t[a].LocalTime end) do

      if not doPurge and v.LocalTime < purgeTime then
        --DebugWrite("DO PURGE: " .. self._settings.Name .. "");
        doPurge = true;
      end

      if not doPurge then
        --DebugWrite("BREAK LOOP: " .. self._settings.Name .. " - v.LocalTime: " .. tostring(v.LocalTime) .. " - purgeTime: " .. purgeTime);
        break;
      end

      --DebugWrite("First Message Time: " .. self._settings.Name .. " (" ..self._settings.ChatGroups[v.ChatGroupId].Name .. ") - " .. tostring(self._settings.ChatGroups[v.ChatGroupId].FirstMessageTime));
      --DebugWrite("        self._settings.ChatGroups[v.ChatGroupId].PurgeCount: " .. tostring(self._settings.ChatGroups[v.ChatGroupId].PurgeCount));
      if self._settings.ChatGroups[v.ChatGroupId].PurgeCount == -1
         and self._settings.ChatGroups[v.ChatGroupId].FirstMessageTime >= 0
         and self._settings.ChatGroups[v.ChatGroupId].FirstMessageTime < purgeTime
      then
        --DebugWrite("CLEAR TEXT: " .. self._settings.Name .. " (" .. self._settings.ChatGroups[v.ChatGroupId].Name .. ")");
        --DebugWrite("          v.ChatGroupId : " .. tostring(v.ChatGroupId) .. " - Id: " .. self._settings.ChatGroups[v.ChatGroupId].Id);
        self._settings.ChatGroups[v.ChatGroupId].PurgeCount = 0;
        self._settings.ChatGroups[v.ChatGroupId].FirstMessageTime = 0;
        self._settings.ChatGroups[v.ChatGroupId].OriginalMessageCount = self._settings.ChatGroups[v.ChatGroupId].MessageCount;
        self._settings.ChatGroups[v.ChatGroupId].MessageCount = 0;
        self:SetText("", v.ChatGroupId);
      end

      if self._settings.ChatGroups[v.ChatGroupId].PurgeCount >= 0 then

        if v.LocalTime < purgeTime then
          --DebugWrite("CLEAR A - " .. self._settings.Name);
          self._entityMessages[k] = nil;
        elseif self._settings.ChatGroups[v.ChatGroupId].OriginalMessageCount - self._settings.ChatGroups[v.ChatGroupId].PurgeCount >= 200 then
          --DebugWrite("CLEAR B - " .. self._settings.Name);
          self._entityMessages[k] = nil;
        else
          if self._settings.ChatGroups[v.ChatGroupId].FirstMessageTime == 0 then
            self._settings.ChatGroups[v.ChatGroupId].FirstMessageTime = v.LocalTime;
          end
          --DebugWrite("APPEND TEXT: " .. self._settings.Name .. " (" ..self._settings.ChatGroups[v.ChatGroupId].Name .. ")");
          self:AppendText(v.StrippedMessage, v.Message, nil, v.ChatGroupId);
        end

        self._settings.ChatGroups[v.ChatGroupId].PurgeCount = self._settings.ChatGroups[v.ChatGroupId].PurgeCount + 1;

      end

    end

    if doPurge then
      for key,value in spairs(self._settings.ChatGroups, function(t,aa,bb) return t[bb].Order > t[aa].Order end) do
        if self._settings.ChatGroups[key].PurgeCount >= 0 then
          self._settings.ChatGroups[key].MessageCount = self._settings.ChatGroups[key].PurgeCount;
          self._settings.ChatGroups[key].PurgeCount = -1;
          self._settings.ChatGroups[key].OriginalMessageCount = 0;
        end
      end

      self.ChatGroups:UpdateGroups();
    end

  end

  function self:ProcessChatMessage(sender, args)

    if self._settings.Enabled then

      local entityName = args.Message:match('<Select:IID:0x%x2%x%x%x%x%x%x%x%x%x%x%x%x%x%x>(.+)<\\Select>');
      local entityType = nil;

      if entityName ~= nil then
        entityType = "Player";
      else
        entityName = args.Message:match('<Select:IID:0x%x3%x%x%x%x%x%x%x%x%x%x%x%x%x%x>(.+)<\\Select>');
        if entityName ~= nil then
          entityType = "NPC";
        else
          entityName = args.Message:match('<Select:IID:0x%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x>(.+)<\\Select>');
          entityType = "Other";
        end
      end

      local chatType = args.ChatType;
      local secondaryChatType = nil;

      --[[
      if chatType == Turbine.ChatType.Tell then
        if false then
          --Turbine.Shell.WriteLine("sender: " .. type(sender));
          for key,value in pairs(sender) do
            Turbine.Shell.WriteLine("key: " .. key);
            Turbine.Shell.WriteLine("type(sender[key]): " .. type(sender[key]));
          end
          for key,value in pairs(args) do
            Turbine.Shell.WriteLine("key: " .. key);
            Turbine.Shell.WriteLine("type(args[key]): " .. type(args[key]));
            Turbine.Shell.WriteLine("args[key]: " .. args[key]);
          end
          Turbine.Shell.WriteLine("args.ChatType: " .. args.ChatType);
          Turbine.Shell.WriteLine("args.Message: " .. args.Message);
        end
        return;
      end ]]

      if chatType == Turbine.ChatType.Standard then
        if string.find(args.Message, " rolled %(") then
          chatType = 400; -- Rolls
          secondaryChatType = 4;
        end
      elseif chatType == Turbine.ChatType.Say then
        if string.sub(args.Message, 1, 9) ~= "You say, " then
          if entityType ~= "Player" then
            chatType = 500; -- Non-player Say
            secondaryChatType = 5;
          end
        end
      end

      for key,value in pairs(self._settings.ChatGroups) do

        if value.Channels[chatType] then

          if chatType == Turbine.ChatType.SelfLoot and value.SelfLootNoCoins then
            if args.Message:match('You looted(.+)coins.') or args.Message:match('You looted(.+)pieces.')then
              return;
            end
          elseif chatType == Turbine.ChatType.Kinship and value.KinNoAnnounce then
            if not string.find(args.Message, ":") then
              if args.Message:match(' has come online.') or args.Message:match(' has gone offline.') then
                return;
              end
            end
          --[[ -- Removed for now, MOTD shows before plugin loads
          elseif chatType == Turbine.ChatType.Kinship and value.KinNoMOTD then
            if not string.find(args.Message, ":") then
              if args.Message:match("Your kinship's Message of the Day") then
                return;
              end
            end
          ]]
          end

          local channelChatColor = value.ChannelColors[chatType];
          if not (channelChatColor) then
            channelChatColor = "#ffffff";
          end

          -- Sent to InteractiveChatText label - Preserves all tags, removes new lines
          local orgmsg = args.Message;
          orgmsg = orgmsg:gsub("[\n\r]", "");

          -- Sent to ChatText label - Strip out all tags, , removes new lines
          local msg = args.Message;
          msg = msg:gsub("%b<>", "");
          msg = msg:gsub("[\n\r]", "");

          if string.len(msg) > 0 and string.gsub(msg, "%s+", "") ~= "" then

            if chatType == Turbine.ChatType.Tell then
              msg = "(someone) tells you, " .. msg;
              orgmsg = "(someone) tells you, " .. orgmsg;
            end

            if value.IncludeTimestamps then
              -- [06/08 02:07:06 AM]
              local dte = Turbine.Engine.GetDate();
              msg = "["..dte.Month.."/"..dte.Day.." "..dte.Hour..":"..dte.Minute..":" .. dte.Second .."] " .. msg;
              orgmsg = "["..dte.Month.."/"..dte.Day.." "..dte.Hour..":"..dte.Minute..":" .. dte.Second .."] " .. orgmsg;
            end

            local secondaryMsg = nil;
            local secondaryOrgmsg = nil;
            if value.Channels[secondaryChatType] then
              local secondaryChannelChatColor = value.ChannelColors[secondaryChatType];
              if not (secondaryChannelChatColor) then
                secondaryChannelChatColor = "#ffffff";
              end
              secondaryMsg = "<rgb=" .. secondaryChannelChatColor .. ">" .. msg .. "</rgb>";
              secondaryOrgmsg = "<rgb=" .. secondaryChannelChatColor .. ">" .. orgmsg .. "</rgb>";
            end

            msg = "<rgb=" .. channelChatColor .. ">" .. msg .. "</rgb>";
            orgmsg = "<rgb=" .. channelChatColor .. ">" .. orgmsg .. "</rgb>";

            -- BEGIN MESSAGE TRACKING ----------

            local date = Turbine.Engine.GetDate();
            local timestamp = date.Year .. "-" .. date.Month .. "-" .. date.Day .. " " .. date.Hour .. ":" .. date.Minute .. ":" .. date.Second;

            value.MessageCount = value.MessageCount + 1;
            if value.FirstMessageTime == -1 then
              value.FirstMessageTime = Turbine.Engine.GetLocalTime();
            end
            table.insert(
              self._entityMessages,
              {
                ChatGroupId = value.Id,
                LocalTime = Turbine.Engine.GetLocalTime(),
                Timestamp = timestamp,
                ChatType = args.ChatType,
                EntityType = entityType,
                EntityName = entityName,
                Message = orgmsg,
                StrippedMessage = msg
              }
            );

            if entityName ~= nil then
              if self._entityMessageCounts[entityName] == nil then
                self._entityMessageCounts[entityName] = 1;
              else
                self._entityMessageCounts[entityName] = self._entityMessageCounts[entityName] + 1;
              end

              if self._entityTypes[entityName] == nil then
                self._entityTypes[entityName] = entityType;
              end
            end

            -- END MESSAGE TRACKING ----------


            -- Send messages to chat labels
            self:AppendText(msg, orgmsg, channelChatColor, value.Id);

            if secondaryMsg ~= nil then
              self:AppendText(secondaryMsg, secondaryOrgmsg, channelChatColor, value.Id);
            end

            -- Check length and update color of Clear button when text count is high
            self.ChatWrapper:UpdateClearButtonColor();

          end
        end
      end

      if self._settings.PurgeChatMinutes > 0 then
        self:DoPurge();
      end

    end

  end

  function self:AppendText(msg, orgmsg, channelChatColor, chatGroupId)

    local chatWindow, interactiveChatWindow = self.ChatWrapper:GetChatWindowsById(chatGroupId);
    if chatWindow:GetTextLength() > 0 then
      chatWindow:AppendText("\n");
      interactiveChatWindow:AppendText("\n");
    end

    chatWindow:AppendText(msg);
    interactiveChatWindow:AppendText(orgmsg);

    --DebugWrite("msg: ".. msg);
    --DebugWrite("chatGroupId: ".. tostring(chatGroupId));

    if self._settings.Enabled and channelChatColor ~= nil then
      if not self._settings.Maximized then
        self.Toolbar:ShowButtons(true);
        self.MinMax:SetForeColor(Turbine.UI.Color(1, self.HexToColor(channelChatColor)));
        self.MinMax2:SetForeColor(self.MinMax:GetForeColor());
      elseif self._chatGroupCount > 1 and chatGroupId ~= self._settings.ActiveChatGroupId then
        self.ChatGroups:SetMissedChat(chatGroupId);
      end
    end

  end

  function self:SetText(msg, chatGroupId)
    --DebugWrite("SetText - chatGroupId: " .. tostring(chatGroupId));
    if chatGroupId == nil then
      local chatGroupWindows = self.ChatWrapper:GetControls();
      for i=1,chatGroupWindows:GetCount() do
        if chatGroupWindows:Get(i).Type == "Wrapper" then
          local chatWindows = chatGroupWindows:Get(i):GetControls();
          for ii=1,chatWindows:GetCount() do
            if chatWindows:Get(ii).Type == "Chat" or chatWindows:Get(ii).Type == "InteractiveChat" then
              chatWindows:Get(ii):SetText(msg);
            end
          end
        end
      end
    else
      local chatWindow, interactiveChatWindow = self.ChatWrapper:GetChatWindowsById(chatGroupId);
      chatWindow:SetText(msg);
      interactiveChatWindow:SetText(msg);
    end
  end

  function self:GetText(msg, chatGroupId)
    local chatWindow, interactiveChatWindow = self.ChatWrapper:GetChatWindowsById(chatGroupId);
    return interactiveChatWindow:GetText();
  end

  function self:GetTextLength(chatGroupId)
    local chatWindow, interactiveChatWindow = self.ChatWrapper:GetChatWindowsById(chatGroupId);
    return interactiveChatWindow:GetTextLength();
  end

  function self:GetEntityMessageCounts(ct)

    local result = "";
    local loopct = 0;
    for k,v in spairs(self._entityMessageCounts, function(t,a,b) return t[b] < t[a] end) do
      result = result .. "<rgb=#cccccc>" .. v .. " - " .. k .. "</rgb>\n";
      loopct = loopct + 1;
      if loopct >= ct then
        break;
      end
    end

    return result;

  end

  function self:BindButtonHoverEvents()
    local buttons = self.Toolbar:GetControls();
    for i=1,buttons:GetCount() do

      ---@diagnostic disable-next-line: duplicate-set-field
      buttons:Get(i).MouseHover = function(sender,args)
        self:OnHover();
        if self._settings.TipsEnabled then
          Libraries.Hint( sender, sender.TipText);
        end
      end

      ---@diagnostic disable-next-line: duplicate-set-field
      buttons:Get(i).MouseLeave = function(sender,args)
        --DebugWrite("buttons:Get(i).MouseLeave");
        self:OnLeave();
        if self._settings.TipsEnabled then
          Libraries.Hint.Remove();
        end
      end
    end

  end

  ------------------------------------------BEGIN DISPLAY INIT------------------------------------------------------------------

  self.ChatWrapper:SetActiveChatGroup(self._settings.ActiveChatGroupId);

  if not self._settings.Maximized then

    self:SetWidth(self.MinMax:GetLeft() + self.MinMax:GetWidth() + 1);
    self:SetHeight(self._headerHeight + 1);

    local windowRight = self._settings.Left + self._settings.Width;
    if self._settings.CollapseRight then
      self:SetLeft(windowRight - self:GetWidth());
    end

    if self._settings.ToolbarBottom then
      self.Toolbar:SetTop(0);
      self.ToolbarBackground:SetTop(self.Toolbar:GetTop());
      self:SetTop((self:GetTop() + self._settings.Height) - self._headerHeight);
    end

  else
    self.Toolbar:UpdateLayout();
  end

  for key,value in spairs(self._settings.ChatGroups, function(t,aa,bb) return t[bb].Order > t[aa].Order end) do
    self._settings.ChatGroups[key].HasUnreadMessages = false;
  end

  if not self._settings.ToolbarPinned then
    self.Toolbar:ShowButtons(false);
  end

  self.Toolbar:MouseEnter({});
  self.Toolbar:MouseLeave({});

  self:BindButtonHoverEvents();

  ------------------------------------------END DISPLAY INIT------------------------------------------------------------------

end